import collections
import os
from itertools import chain
from typing import Dict, Set, Any, List, Union, Optional

from .ExternalLink import ExternalLink
from .Link import Link
from .Machine import Machine
from .. import utils


class Lab(object):
    """A Kathara network scenario, containing information about devices and collision domains.

    Attributes:
        name (str): The name of the network scenario.
        description (str): A short description of the network scenario.
        version (str): The version of the network scenario.
        author (str): The author of the network scenario.
        email (str): The email of the author of the network scenario.
        web (str): The web address of the author of the network scenario.
        path (str): The path of the network scenario, if exists.
        hash (str): The hash identifier of the network scenario.
        machines (Dict[str, Kathara.model.Machine]): The devices of the network scenario. Keys are device names, Values
            are Kathara device objects.
        links (Dict[str, Kathara.model.Link]): The collision domains of the network scenario.
            Keys are collision domains names, Values are Kathara collision domain objects.
        general_options (Dict[str, Any]): Keys are option names, values are option values.
        has_dependencies (bool): True if there are dependencies among the devices boot.
        shared_startup_path(str) The path of the shared startup file, if exists.
        shared_shutdown_path(str) The path of the shared shutdown file, if exists.
        shared_folder(str) The path of the shared folder, if exists.
    """
    __slots__ = ['name', 'description', 'version', 'author', 'email', 'web',
                 'path', 'hash', 'machines', 'links', 'general_options', 'has_dependencies',
                 'shared_startup_path', 'shared_shutdown_path', 'shared_folder']

    def __init__(self, name: Optional[str], path: str = None) -> None:
        """Create a new instance of a Kathara network scenario.

        Args:
            name (str): The name of the network scenario.
            path (str): The path to the network scenario directory, if exists.

        Returns:
            None
        """
        self.name: Optional[str] = name
        self.description: Optional[str] = None
        self.version: Optional[str] = None
        self.author: Optional[str] = None
        self.email: Optional[str] = None
        self.web: Optional[str] = None

        self.machines: Dict[str, Machine] = {}
        self.links: Dict[str, Link] = {}

        self.general_options: Dict[str, Any] = {}

        self.has_dependencies: bool = False

        self.path: str = path
        self.shared_startup_path: Optional[str] = None
        self.shared_shutdown_path: Optional[str] = None

        self.hash: str = ""

        if self.path:
            self.path = path
            self.hash = utils.generate_urlsafe_hash(self.path)
            shared_startup_file = os.path.join(self.path, 'shared.startup')
            self.shared_startup_path = shared_startup_file if os.path.exists(shared_startup_file) else None

            shared_shutdown_file = os.path.join(self.path, 'shared.shutdown')
            self.shared_shutdown_path = shared_shutdown_file if os.path.exists(shared_shutdown_file) else None
        else:
            self.hash = utils.generate_urlsafe_hash(self.name)

        self.shared_folder: Optional[str] = None

    def connect_machine_to_link(self, machine_name: str, link_name: str, machine_iface_number: int = None) -> None:
        """Connect the specified device to the specified collision domain.

        Args:
            machine_name (str): The device name.
            link_name (str): The collision domain name.
            machine_iface_number (int): The number of the device interface to connect. If it is None, the first free
                number is used.

        Returns:
            None

        Raises:
            Exception: If an already used interface number is specified.
        """
        machine = self.get_or_new_machine(machine_name)
        link = self.get_or_new_link(link_name)

        machine.add_interface(link, number=machine_iface_number)

    def assign_meta_to_machine(self, machine_name: str, meta_name: str, meta_value: str) -> None:
        """Assign a meta information to the specified device.

        Args:
            machine_name (str): The name of the device.
            meta_name (str): The name of the meta property.
            meta_value (str): The value of the meta property.

        Returns:
            None

        Raises:
            MachineOptionError: If invalid values are specified for meta properties.
        """
        machine = self.get_or_new_machine(machine_name)

        machine.add_meta(meta_name, meta_value)

    def attach_external_links(self, external_links: Dict[str, List[ExternalLink]]) -> None:
        """Attach external collision domains to the network scenario.

        Args:
            external_links (Dict[Kathara.model.Link, List[Kathara.model.ExternalLink]]): Keys are Link objects,
            values are ExternalLink objects.

        Returns:
            None
        """
        for (link_name, link_external_links) in external_links.items():
            if link_name not in self.links:
                raise Exception("Collision domain `%s` (declared in lab.ext) not found in lab "
                                "collision domains." % link_name)

            self.links[link_name].external += link_external_links

    def check_integrity(self) -> None:
        """Check if the network interfaces numbers of all the devices in the network scenario are correctly assigned."""
        for machine in self.machines:
            self.machines[machine].check()

    def get_links_from_machines(self, selected_machines: Union[List[str], Set[str]]) -> Dict[str, Link]:
        """Return the name of the links connected to the selected_machines.

        Args:
            selected_machines (Set[str]): A set with selected devices names.

        Returns:
            Dict[str, Link]: Keys are Link names, values are Link objects.
        """
        # Intersect selected machines names with self.machines keys
        selected_machines = set(self.machines.keys()) & set(selected_machines)
        # Apply filtering
        machines = [v for (k, v) in self.machines.items() if k in selected_machines]

        # Get only selected machines Link objects.
        selected_links = set(chain.from_iterable([machine.interfaces.values() for machine in machines]))
        selected_links = {link.name: link for link in selected_links}

        return selected_links

    def apply_dependencies(self, dependencies: List[str]) -> None:
        """Order the list of devices of the network scenario to satisfy the boot dependencies.

        Args:
            dependencies (List[str]): If not empty, dependencies are applied.

        Returns:
            None
        """

        def dep_sort(item: str) -> int:
            try:
                return dependencies.index(item) + 1
            except ValueError:
                return 0

        self.machines = collections.OrderedDict(sorted(self.machines.items(), key=lambda t: dep_sort(t[0])))
        self.has_dependencies = True

    def get_or_new_machine(self, name: str, **kwargs: Dict[str, Any]) -> Machine:
        """Get the specified device. If it not exists, create and add it to the devices list.

        Args:
            name (str): The name of the device
            **kwargs (Dict[str, Any]): Contains device meta information.
                Keys are meta property names, values are meta property values.

        Returns:
            Kathara.model.Machine: A Kathara device.
        """
        if name not in self.machines:
            self.machines[name] = Machine(self, name, **kwargs)

        return self.machines[name]

    def get_or_new_link(self, name: str) -> Link:
        """Get the specified collision domain. If it not exists, create and add it to the collision domains list.

        Args:
            name (str): The name of the collision domain.

        Returns:
            Kathara.model.Link: A Kathara collision domain.
        """
        if name not in self.links:
            self.links[name] = Link(self, name)

        return self.links[name]

    def create_shared_folder(self) -> None:
        """If the network scenario has a directory, create the network scenario shared folder.

        Returns:
            None

        Raises:
            Exception: The shared folder is a Symlink, delete it.
            OSError: Permission error.
        """
        if not self.has_path():
            return
        try:
            self.shared_folder = os.path.join(self.path, 'shared')
            if not os.path.isdir(self.shared_folder):
                os.mkdir(self.shared_folder)
            elif os.path.islink(self.shared_folder):
                raise Exception("`shared` folder is a symlink, delete it.")
        except OSError:
            # Do not create shared folder if not permitted.
            return

    def has_path(self) -> bool:
        """Check if the network scenario has a directory.

        Returns:
            bool: True if it self.path is not None, else False.
        """
        return self.path is not None

    def add_option(self, name: str, value: Any) -> None:
        """Add an option to the network scenario.

        Args:
            name (str): The name of the option.
            value (Any): The value of the option.

        Returns:
            None
        """
        if value is not None:
            self.general_options[name] = value

    def __repr__(self) -> str:
        return "Lab(%s, %s, %s, %s)" % (self.path, self.hash, self.machines, self.links)

    def __str__(self) -> str:
        lab_info = ""

        if self.name:
            lab_info += "Name: %s\n" % self.name

        if self.description:
            lab_info += "Description: %s\n" % self.description

        if self.version:
            lab_info += "Version: %s\n" % self.version

        if self.author:
            lab_info += "Author(s): %s\n" % self.author

        if self.email:
            lab_info += "Email: %s\n" % self.email

        if self.web:
            lab_info += "Website: %s" % self.web

        return lab_info



from __future__ import annotations

from typing import Any


class CliArgs(object):
    __slots__ = ['args']

    __instance: CliArgs = None

    @staticmethod
    def get_instance() -> CliArgs:
        if CliArgs.__instance is None:
            CliArgs()

        return CliArgs.__instance

    def __init__(self) -> None:
        if CliArgs.__instance is not None:
            raise Exception("This class is a singleton!")
        else:
            self.args = {}

            CliArgs.__instance = self

    def __getattr__(self, item: str) -> Any:
        if item in self.args:
            return self.args[item]
        else:
            return None



import logging
from typing import Callable

from docker import DockerClient
from docker.errors import NotFound

from ... import utils
from ...os.Networking import Networking
from ...setting.Setting import Setting

PLUGIN_NAME = "kathara/katharanp:" + utils.get_architecture()
XTABLES_CONFIGURATION_KEY = "xtables_lock"
XTABLES_LOCK_PATH = "/run/xtables.lock"


class DockerPlugin(object):
    """Class responsible for interacting with Docker Plugins."""
    __slots__ = ['client']

    def __init__(self, client: DockerClient):
        self.client: DockerClient = client

    def check_and_download_plugin(self) -> None:
        """Check the presence of the Kathara Network Plugin and download it or upgrade it, if needed.

        Returns:
            None
        """

        try:
            logging.debug("Checking plugin `%s`..." % PLUGIN_NAME)
            plugin = self.client.plugins.get(PLUGIN_NAME)
            # Check for plugin updates.
            plugin.upgrade()
        except NotFound:
            if Setting.get_instance().remote_url is None:
                logging.info("Installing Kathara Network Plugin...")
                plugin = self.client.plugins.install(PLUGIN_NAME)
                logging.info("Kathara Network Plugin installed successfully!")
            else:
                raise Exception("Kathara Network Plugin not found on remote Docker connection.")

        if Setting.get_instance().remote_url is None:
            xtables_lock_mount = self._get_xtables_lock_mount()
            if not plugin.enabled:
                self._configure_xtables_mount(plugin, xtables_lock_mount)
                logging.debug("Enabling plugin `%s`..." % PLUGIN_NAME)
                plugin.enable()
            else:
                # Get the mount of xtables.lock from the current plugin configuration
                mount_obj = list(filter(lambda x: x["Name"] == XTABLES_CONFIGURATION_KEY,
                                        plugin.attrs["Settings"]["Mounts"])).pop()

                # If it's not equal to the computed one, fix the mount
                if mount_obj["Source"] != xtables_lock_mount:
                    plugin.disable()
                    self._configure_xtables_mount(plugin, xtables_lock_mount)
                    logging.debug("Enabling plugin `%s`..." % PLUGIN_NAME)
                    plugin.enable()
        else:
            if not plugin.enabled:
                raise Exception("Kathara Network Plugin not enabled on remote Docker connection.")

    def _get_xtables_lock_mount(self) -> Callable:
        def _mount_xtables_lock_linux():
            iptables_version = Networking.get_iptables_version()
            return "" if 'nf_tables' in iptables_version else XTABLES_LOCK_PATH

        def _mount_xtables_lock_windows():
            docker_info = self.client.info()
            return "" if 'microsoft' not in docker_info['KernelVersion'] else XTABLES_LOCK_PATH

        return utils.exec_by_platform(_mount_xtables_lock_linux, _mount_xtables_lock_windows, lambda: "")

    @staticmethod
    def _configure_xtables_mount(plugin, xtables_lock_mount) -> None:
        logging.debug("Configuring xtables.lock source to `%s`..." % xtables_lock_mount)
        plugin.configure({
            XTABLES_CONFIGURATION_KEY + '.source': xtables_lock_mount
        })



import argparse
from typing import List

from ... import utils
from ...foundation.cli.command.Command import Command
from ...manager.Kathara import Kathara
from ...strings import strings, wiki_description
from ...trdparty.curses.curses import Curses


class ListCommand(Command):
    def __init__(self) -> None:
        Command.__init__(self)

        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(
            prog='kathara list',
            description=strings['list'],
            epilog=wiki_description,
            add_help=False
        )

        self.parser.add_argument(
            '-h', '--help',
            action='help',
            default=argparse.SUPPRESS,
            help='Show an help message and exit.'
        )

        self.parser.add_argument(
            '-a', '--all',
            required=False,
            action='store_true',
            help='Show all running Kathara devices of all users. MUST BE ROOT FOR THIS OPTION.'
        )

        self.parser.add_argument(
            '-l', '--live',
            required=False,
            action='store_true',
            help='Live mode.'
        )

        self.parser.add_argument(
            '-n', '--name',
            metavar='DEVICE_NAME',
            required=False,
            help='Show only information about a specified device.'
        )

    def run(self, current_path: str, argv: List[str]) -> None:
        self.parse_args(argv)
        args = self.get_args()

        if args['all'] and not utils.is_admin():
            raise Exception("You must be root in order to show all Kathara devices of all users.")

        all_users = bool(args['all'])

        if args['live']:
            if args['name']:
                self._get_machine_live_info(args['name'], all_users)
            else:
                self._get_lab_live_info(all_users)
        else:
            if args['name']:
                print(Kathara.get_instance().get_formatted_machine_info(args['name'], all_users=all_users))
            else:
                lab_info = Kathara.get_instance().get_formatted_lab_info(all_users=all_users)

                print(next(lab_info))

    @staticmethod
    def _get_machine_live_info(machine_name: str, all_users: bool) -> None:
        Curses.get_instance().init_window()

        try:
            while True:
                Curses.get_instance().print_string(
                    Kathara.get_instance().get_formatted_machine_info(machine_name, all_users=all_users)
                )
        finally:
            Curses.get_instance().close()

    @staticmethod
    def _get_lab_live_info(all_users: bool) -> None:
        lab_info = Kathara.get_instance().get_formatted_lab_info(all_users=all_users)

        Curses.get_instance().init_window()

        try:
            while True:
                Curses.get_instance().print_string(next(lab_info))
        except StopIteration:
            pass
        finally:
            Curses.get_instance().close()



import argparse
import logging
import os
import shutil
import sys
import time
from typing import List

from .LcleanCommand import LcleanCommand
from .LstartCommand import LstartCommand
from ... import utils
from ...exceptions import TestError
from ...foundation.cli.command.Command import Command
from ...strings import strings, wiki_description
from ...test.BuiltinTest import BuiltInTest
from ...test.UserTest import UserTest


class LtestCommand(Command):
    def __init__(self) -> None:
        Command.__init__(self)

        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(
            prog='kathara ltest',
            description=strings['ltest'],
            epilog=wiki_description,
            add_help=False
        )

        self.parser.add_argument(
            '-h', '--help',
            action='help',
            default=argparse.SUPPRESS,
            help='Show an help message and exit.'
        )

        self.parser.add_argument(
            '-d', '--directory',
            required=False,
            help='Specify the folder containing the lab.'
        )
        self.parser.add_argument(
            '-R', '--rebuild-signature',
            dest="rebuild_signature",
            required=False,
            action='store_true',
            help='Force generating a new signature for the lab, even if one already exists. '
                 'Overwrites any existing signature.'
        )
        self.parser.add_argument(
            '--wait',
            required=False,
            metavar='MINUTES',
            help='Minutes to wait from lab startup before running the tests (can be a decimal number).'
        )
        self.parser.add_argument(
            '--verify',
            required=False,
            choices=['builtin', 'user', 'both'],
            help='Compares current lab state with stored signature.'
        )

    def run(self, current_path: str, argv: List[str]) -> None:
        self.parse_args(argv)
        args = self.get_args()

        lab_path = args['directory'].replace('"', '').replace("'", '') if args['directory'] else current_path
        lab_path = utils.get_absolute_path(lab_path)

        signature_test_path = None
        if not args['verify']:
            signature_test_path = os.path.join(lab_path, "_test", "signature")

            if os.path.exists(signature_test_path) and not args['rebuild_signature']:
                logging.error("Signature for current lab already exists. Exiting...")
                sys.exit(1)

        # Tests run without terminals, no shared and /hosthome dirs.
        new_argv = ["--noterminals", "--no-shared", "--no-hosthome"]

        # Start the lab
        lab = LstartCommand().run(lab_path, new_argv)

        if args['wait']:
            try:
                sleep_minutes = float(args['wait'])
                if sleep_minutes < 0:
                    raise ValueError()

                logging.info("Waiting %s minutes before running tests..." % sleep_minutes)
                time.sleep(sleep_minutes * 60)
            except ValueError:
                raise ValueError("--wait value is not valid!")

        # Run Tests and store in signature
        builtin_test = BuiltInTest(lab)
        user_test = UserTest(lab)

        builtin_test_passed = True
        user_test_passed = True

        if not args['verify']:
            if not os.path.exists(signature_test_path) or args['rebuild_signature']:
                shutil.rmtree(signature_test_path, ignore_errors=True)
                os.makedirs(signature_test_path, exist_ok=True)

                builtin_test.create_signature()
                user_test.create_signature()
        else:
            result_test_path = os.path.join(lab.path, "_test", "results")

            shutil.rmtree(result_test_path, ignore_errors=True)
            os.makedirs(result_test_path, exist_ok=True)

            try:
                if args['verify'] == "builtin" or args['verify'] == "both":
                    builtin_test_passed = builtin_test.test()
                if args['verify'] == "user" or args['verify'] == "both":
                    user_test_passed = user_test.test()
            except TestError as e:
                logging.error(str(e))

        # Clean the lab at the end of the test.
        LcleanCommand().run(lab_path, [])

        if args['verify']:
            if not builtin_test_passed or not user_test_passed:
                sys.exit(1)



import argparse
import logging
import os
import platform
import sys
from typing import List

from ... import utils
from ... import version
from ...foundation.cli.command.Command import Command
from ...manager.Kathara import Kathara
from ...model.Lab import Lab
from ...setting.Setting import Setting
from ...strings import strings, wiki_description


class CheckCommand(Command):
    def __init__(self) -> None:
        Command.__init__(self)

        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(
            prog='kathara check',
            description=strings['check'],
            epilog=wiki_description,
            add_help=False
        )

        self.parser.add_argument(
            '-h', '--help',
            action='help',
            default=argparse.SUPPRESS,
            help='Show an help message and exit.'
        )

    def run(self, current_path: str, argv: List[str]) -> None:
        self.parse_args(argv)
        args = self.get_args()

        print("*\tCurrent Manager is: %s" % Kathara.get_instance().get_formatted_manager_name())

        print("*\tManager version is: %s" % Kathara.get_instance().get_release_version())

        print("*\tPython version is: %s" % sys.version.replace("\n", "- "))

        print("*\tKathara version is: %s" % version.CURRENT_VERSION)

        def linux_platform_info():
            info = os.uname()
            return "%s-%s-%s" % (info.sysname, info.release, info.machine)

        platform_info = utils.exec_by_platform(
            linux_platform_info, lambda: platform.platform(), lambda: platform.platform()
        )
        print("*\tOperating System version is: %s" % str(platform_info))

        print("*\tTrying to run `Hello World` container...")

        Setting.get_instance().open_terminals = False
        args['no_shared'] = False
        args['no_hosthome'] = False

        lab = Lab("kathara_vlab")
        lab.get_or_new_machine("hello_world")

        try:
            Kathara.get_instance().deploy_lab(lab)
            print("*\tContainer run successfully.")
            Kathara.get_instance().undeploy_lab(lab.hash)
        except Exception as e:
            logging.exception("\t! Running `Hello World` failed: %s" % str(e))



import difflib
import logging
import os
from typing import Union, Dict, List, Optional

from deepdiff import DeepDiff

from .. import utils
from ..exceptions import MachineSignatureNotFoundError
from ..foundation.test.Test import Test
from ..manager.Kathara import Kathara
from ..model.Lab import Lab
from ..model.Machine import Machine
from ..setting.Setting import Setting


class UserTest(Test):
    def __init__(self, lab: Lab) -> None:
        Test.__init__(self, lab)

    def create_signature(self) -> None:
        for (machine_name, machine) in self.lab.machines.items():
            machine_test_file = self._copy_machine_test_file(machine)

            if machine_test_file:
                logging.info("Building `user` signature for device %s..." % machine_name)

                machine_state = self._run_machine_test_file(machine)

                # Write the signature into the proper file
                with open("%s/%s.user" % (self.signature_path, machine_name), 'w') as machine_signature_file:
                    machine_signature_file.write(machine_state)

    def test(self) -> bool:
        test_passed = True

        for (machine_name, machine) in self.lab.machines.items():
            machine_test_file = self._copy_machine_test_file(machine)

            if machine_test_file:
                logging.info("Executing `user` tests for device %s..." % machine_name)

                machine_state = self._run_machine_test_file(machine)

                # Read the signature from machine file
                machine_signature_path = "%s/%s.user" % (self.signature_path, machine.name)
                if os.path.exists(machine_signature_path):
                    with open(machine_signature_path, 'r') as machine_signature_file:
                        machine_signature = machine_signature_file.read()
                else:
                    raise MachineSignatureNotFoundError("Signature for device `%s` not found! Exiting..." %
                                                        machine_name
                                                        )

                # Save machine state into result file
                machine_result_path = "%s/%s.user" % (self.results_path, machine.name)
                with open(machine_result_path, 'w') as machine_result_file:
                    machine_result_file.write(machine_state)

                diff = self.check_signature(machine_signature, machine_state)
                test_passed = False if diff else test_passed

                machine_diff_path = "%s/%s.diff" % (self.results_path, machine.name)
                with open(machine_diff_path, 'a') as machine_diff_file:
                    machine_diff_file.write(utils.format_headers("User Test Result") + '\n')
                    machine_diff_file.write("\n".join(diff) + "\n" if diff else "OK\n")
                    machine_diff_file.write(utils.format_headers() + "\n\n")

        return test_passed

    def check_signature(self, signature: Union[Dict, str], status: Union[Dict, str]) -> Union[DeepDiff, List]:
        signature = signature.splitlines()
        status = status.splitlines()

        # Do the diff between the arrays, n=0 removes context strings
        diff = difflib.unified_diff(signature, status, n=0, lineterm="")
        # Remove headers of the diff from the result
        return [x for x in filter(lambda x: not x.startswith(('---', '+++', '@@')), diff)]

    def _copy_machine_test_file(self, machine: Machine) -> Optional[str]:
        machine_test_file = os.path.join(self.test_path, "%s.test" % machine.name)

        if os.path.exists(machine_test_file):
            Kathara.get_instance().copy_files(machine, {"/%s.test" % machine.name: machine_test_file})

            return machine_test_file
        else:
            return None

    @staticmethod
    def _run_machine_test_file(machine: Machine) -> str:
        # Give execution permissions to test file
        Test._get_machine_command_output(lab_hash=machine.lab.hash,
                                         machine_name=machine.name,
                                         command="chmod u+x /%s.test" % machine.name
                                         )

        # Run the test file inside the container
        (stdout, stderr) = Test._get_machine_command_output(lab_hash=machine.lab.hash,
                                                            machine_name=machine.name,
                                                            command="%s -c /%s.test" % (
                                                                Setting.get_instance().device_shell,
                                                                machine.name
                                                            ))

        return stdout if stdout else stderr if stderr else ""



class MenuPadding(object):
    """
    Class for menu padding. Padding is the area between the menu border and the content of the menu.

    Args:
        top (int): The top padding.
        left (int):  The left padding.
        bottom (int): The bottom padding.
        right (int): The right padding.
    """

    def __init__(self, top=1, left=2, bottom=1, right=2):
        self.__left = left
        self.__right = right
        self.__top = top
        self.__bottom = bottom

    @property
    def left(self):
        """
        The left padding.

        Returns:
            int: The left padding.
        """
        return self.__left

    @left.setter
    def left(self, left):
        self.__left = left

    @property
    def right(self):
        """
        The right padding.

        Returns:
            int: The right padding.
        """
        return self.__right

    @right.setter
    def right(self, right):
        self.__right = right

    @property
    def top(self):
        """
        The top padding.

        Returns:
            int: The top padding.
        """
        return self.__top

    @top.setter
    def top(self, top):
        self.__top = top

    @property
    def bottom(self):
        """
        The bottom padding.

        Returns:
            int: The bottom padding.
        """
        return self.__bottom

    @bottom.setter
    def bottom(self, bottom):
        self.__bottom = bottom



# Generic Exceptions
class ClassNotFoundError(Exception):
    pass


class HTTPConnectionError(Exception):
    pass


class SettingsError(Exception):
    def __init__(self, message) -> None:
        super().__init__("Settings file is not valid: %s\nFix it or delete it before launching." % message)


class DockerDaemonConnectionError(Exception):
    pass


class NotSupportedError(Exception):
    def __init__(self, message) -> None:
        super().__init__("Not Supported: %s" % message)


# OS Exceptions
class PrivilegeError(Exception):
    pass


# Lab Exceptions
class LabAlreadyExistsError(Exception):
    pass


# Machine Exceptions
class MountDeniedError(Exception):
    pass


class MachineAlreadyExistsError(Exception):
    pass


class NonSequentialMachineInterfaceError(Exception):
    pass


class MachineOptionError(Exception):
    pass


class MachineCollisionDomainConflictError(Exception):
    pass


# Test Exceptions
class TestError(Exception):
    pass


class MachineSignatureNotFoundError(TestError):
    pass



from typing import List, Any

from . import Lab as LabPackage
from .ExternalLink import ExternalLink

BRIDGE_LINK_NAME = "kathara_host_bridge"


class Link(object):
    """A Kathara collision domain.

    Contains information about the collision domain and the API object to interact with the Manager.

    Attributes:
        lab (Kathara.model.Lab): The Kathara network Scenario of the collision domain.
        name (str): The name of the collision domain.
        external (List[Kathara.model.ExternalLink]): List of the ExternalLinks attached to this collision domain.
        api_object (Any): To interact with the current Kathara Manager.
    """
    __slots__ = ['lab', 'name', 'external', 'api_object']

    def __init__(self, lab: 'LabPackage.Lab', name) -> None:
        self.lab: 'LabPackage.Lab' = lab
        self.name: str = name
        self.external: List[ExternalLink] = []
        self.api_object: Any = None

    def __repr__(self) -> str:
        return "Link(%s, %s)" % (self.name, self.external)



import argparse
import sys
from typing import List

from ... import utils
from ...foundation.cli.command.Command import Command
from ...manager.Kathara import Kathara
from ...strings import strings, wiki_description


class ExecCommand(Command):
    def __init__(self) -> None:
        Command.__init__(self)

        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(
            prog='kathara exec',
            description=strings['exec'],
            epilog=wiki_description,
            add_help=False
        )

        self.parser.add_argument(
            '-h', '--help',
            action='help',
            default=argparse.SUPPRESS,
            help='Show an help message and exit.'
        )

        group = self.parser.add_mutually_exclusive_group(required=False)

        group.add_argument(
            '-d', '--directory',
            help='Specify the folder containing the lab.',
        )
        group.add_argument(
            '-v', '--vmachine',
            dest="vmachine",
            action="store_true",
            help='The device has been started with vstart command.',
        )
        self.parser.add_argument(
            '--no-stdout',
            dest="no_stdout",
            action="store_true",
            help='Disable stdout of the executed command.',
        )
        self.parser.add_argument(
            '--no-stderr',
            dest="no_stderr",
            action="store_true",
            help='Disable stderr of the executed command.',
        )
        self.parser.add_argument(
            'machine_name',
            metavar='DEVICE_NAME',
            help='Name of the device to execute the command into.'
        )
        self.parser.add_argument(
            'command',
            metavar='COMMAND',
            nargs='+',
            help='Shell command that will be executed inside the device.'
        )

    def run(self, current_path: str, argv: List[str]) -> None:
        self.parse_args(argv)
        args = self.get_args()

        if args['vmachine']:
            lab_path = "kathara_vlab"
        else:
            lab_path = args['directory'].replace('"', '').replace("'", '') if args['directory'] else current_path
            lab_path = utils.get_absolute_path(lab_path)

        lab_hash = utils.generate_urlsafe_hash(lab_path)
        exec_output = Kathara.get_instance().exec(lab_hash, args['machine_name'], args['command'])

        try:
            while True:
                (stdout, stderr) = next(exec_output)
                stdout = stdout.decode('utf-8') if stdout else ""
                stderr = stderr.decode('utf-8') if stderr else ""

                if not args['no_stdout']:
                    sys.stdout.write(stdout)
                if stderr and not args['no_stderr']:
                    sys.stderr.write(stderr)
        except StopIteration:
            pass



import mmap
import os
import re

from ...model.Lab import Lab
from ...utils import RESERVED_MACHINE_NAMES


class LabParser(object):
    """
    Class responsible for parsing the lab.conf file.
    """
    @staticmethod
    def parse(path: str) -> Lab:
        """
        Parse the lab.conf and return the corresponding Kathara network scenario.

        Args:
            path (str): The path to lab.conf file.

        Returns:
            Kathara.model.Lab.Lab: A Kathara network scenario.
        """
        lab_conf_path = os.path.join(path, 'lab.conf')

        if not os.path.exists(lab_conf_path):
            raise IOError("No lab.conf in given directory.\n")

        if os.stat(lab_conf_path).st_size == 0:
            raise IOError("lab.conf file is empty.\n")

        # Reads lab.conf in memory so it is faster.
        try:
            with open(lab_conf_path, 'r') as lab_file:
                lab_mem_file = mmap.mmap(lab_file.fileno(), 0, access=mmap.ACCESS_READ)
        except Exception:
            raise IOError("Cannot open lab.conf file.")

        lab = Lab(None, path=path)

        line_number = 1
        line = lab_mem_file.readline().decode('utf-8')
        while line:
            matches = re.search(r"^(?P<key>[a-z0-9_]{1,30})\[(?P<arg>\w+)\]=(?P<value>\".+\"|\'.+\'|\w+)$",
                                line.strip()
                                )

            if matches:
                key = matches.group("key").strip()
                arg = matches.group("arg").strip()
                value = matches.group("value").replace('"', '').replace("'", '')

                if key in RESERVED_MACHINE_NAMES:
                    raise Exception("[ERROR] In line %d: "
                                    "`%s` is a reserved name, you can not use it for a device." % (line_number, key))

                try:
                    # It's an interface, handle it.
                    interface_number = int(arg)

                    if re.search(r"^\w+$", value):
                        lab.connect_machine_to_link(key, value, machine_iface_number=interface_number)
                    else:
                        raise Exception("[ERROR] In line %d: "
                                        "Collision domain `%s` contains non-alphanumeric characters." % (line_number,
                                                                                                         value))
                except ValueError:
                    # Not an interface, add it to the machine metas.
                    lab.assign_meta_to_machine(key, arg, value)
            else:
                if not line.startswith('#') and \
                        line.strip():
                    if not line.startswith("LAB_NAME=") and \
                            not line.startswith("LAB_DESCRIPTION=") and \
                            not line.startswith("LAB_VERSION=") and \
                            not line.startswith("LAB_AUTHOR=") and \
                            not line.startswith("LAB_EMAIL=") and \
                            not line.startswith("LAB_WEB="):
                        raise Exception("[ERROR] In line %d: Invalid characters `%s`." % (line_number, line))
                    else:
                        (key, value) = line.split("=")
                        key = key.replace("LAB_", "").lower()
                        setattr(lab, key, value.replace('"', '').replace("'", '').strip())

            line_number += 1
            line = lab_mem_file.readline().decode('utf-8')

        lab.check_integrity()

        return lab



import collections
import logging
import os
import re
import shutil
import subprocess
import sys
import tarfile
import tempfile
from distutils.util import strtobool
from pathlib import Path
from typing import Dict, Any, Tuple, Optional

from . import Lab as LabPackage
from . import Link as LinkPackage
from .. import utils
from ..exceptions import NonSequentialMachineInterfaceError, MachineOptionError, MachineCollisionDomainConflictError
from ..setting.Setting import Setting


class Machine(object):
    """A Kathara device.

    Contains information about the device and the API object to interact with the Manager.

    Attributes:
        lab (Kathara.model.Lab): The Kathara network Scenario of the device.
        name (str): The name of the device.
        interfaces (List[Kathara.model.Link]): A list of the collision domains of the device.
        meta (Dict[str, Any]): Keys are meta properties name, values are meta properties values.
        startup_commands (List[str]): A list of commands to execute at the device startup.
        api_object (Any): To interact with the current Kathara Manager.
        capabilities (List[str]): The selected capabilities for the device.
        startup_path (str): The path of the device startup file, if exists.
        shutdown_path (str): The path of the device shutdown file, if exists.
        folder (str): The path of the device folder, if exists.
    """
    __slots__ = ['lab', 'name', 'interfaces', 'meta', 'startup_commands', 'api_object', 'capabilities',
                 'startup_path', 'shutdown_path', 'folder']

    def __init__(self, lab: 'LabPackage.Lab', name: str, **kwargs) -> None:
        """Create a new instance of a Kathara device.

        Args:
            lab (Kathara.model.Lab): The Kathara network scenario of the new device.
            name (str): The name of the device.
            **kwargs (Dict[str, Any]): Specifies the optional parameters of the device.

        Returns:
            None
        """
        name = name.strip()
        matches = re.search(r"^[a-z0-9_]{1,30}$", name)
        if not matches:
            raise Exception("Invalid device name `%s`." % name)

        self.lab: LabPackage.Lab = lab
        self.name = name

        self.interfaces = {}

        self.meta: Dict[str, Any] = {
            'sysctls': {},
            'bridged': False,
            'ports': {}
        }

        self.startup_commands = []

        self.api_object = None

        self.capabilities = ["NET_ADMIN", "NET_RAW", "NET_BROADCAST", "NET_BIND_SERVICE", "SYS_ADMIN"]

        self.startup_path = None
        self.shutdown_path = None
        self.folder = None

        if lab.has_path():
            startup_file = os.path.join(lab.path, '%s.startup' % self.name)
            self.startup_path = startup_file if os.path.exists(startup_file) else None

            shutdown_file = os.path.join(lab.path, '%s.shutdown' % self.name)
            self.shutdown_path = shutdown_file if os.path.exists(shutdown_file) else None

            machine_folder = os.path.join(lab.path, '%s' % self.name)
            self.folder = machine_folder if os.path.isdir(machine_folder) else None

        self.update_meta(kwargs)

    def add_interface(self, link: 'LinkPackage.Link', number: int = None) -> None:
        """Add an interface to the device attached to the specified collision domain.

        Args:
            link (Kathara.model.Link): The Kathara collision domain to attach.
            number (int): The number of the new interface. If it is None, the first free number is selected.

        Returns:
            None

        Raises:
            Exception: The interface number specified is already used on the device.
        """
        if number is None:
            number = len(self.interfaces.keys())

        if number in self.interfaces:
            raise Exception("Interface %d already set on device `%s`." % (number, self.name))

        if any(x.name == link.name for x in self.interfaces.values()):
            raise MachineCollisionDomainConflictError(
                "Device `%s` is already connected to collision domain `%s`" % (self.name, link.name)
            )

        self.interfaces[number] = link

    def add_meta(self, name: str, value: Any) -> None:
        """Add a meta property to the device.

        Args:
            name (str): The name of the property.
            value (Any): The value of the property.

        Returns:
            None

        Raises:
            MachineOptionError: The specified value is not valid for the specified property.
        """
        if name == "exec":
            self.startup_commands.append(value)
            return

        if name == "bridged":
            self.meta[name] = bool(strtobool(str(value)))
            return

        if name == "sysctl":
            # Check for valid kv-pair
            if '=' in value:
                (key, val) = value.split('=')
                key = key.strip()
                val = val.strip()
                # Only allow `net.` namespace
                if key.startswith('net.'):
                    # Convert to int if possible
                    self.meta['sysctls'][key] = int(val) if val.isdigit() else val
                else:
                    raise MachineOptionError(
                        "Invalid sysctl value (`%s`) on `%s`, only `net.` namespace is allowed." % (value, self.name)
                    )
            else:
                raise MachineOptionError("Invalid sysctl value (`%s`) on `%s`, missing `=`." % (value, self.name))
            return

        if name == "port":
            if '/' in value:
                (ports, protocol) = value.split('/')
            else:
                (ports, protocol) = value, 'tcp'

            if ':' not in ports:
                host_port, guest_port = 3000, ports
            else:
                (host_port, guest_port) = ports.split(':')

            protocol = protocol.lower()
            if protocol not in ['tcp', 'udp', 'sctp']:
                raise MachineOptionError("Port protocol value not valid on `%s`." % self.name)

            try:
                self.meta['ports'][(int(host_port), protocol)] = int(guest_port)
            except ValueError:
                raise MachineOptionError("Port value not valid on `%s`." % self.name)
            return

        self.meta[name] = value

    def check(self) -> None:
        """Sorts interfaces check if there are missing positions.

        Returns:
            None

        Raises:
            NonSequentialMachineInterfaceError: If there is a missing interface number.
        """
        sorted_interfaces = sorted(self.interfaces.items(), key=lambda kv: kv[0])

        logging.debug("`%s` interfaces are %s." % (self.name, sorted_interfaces))

        for i, (num_iface, _) in enumerate(sorted_interfaces):
            if num_iface != i:
                raise NonSequentialMachineInterfaceError("Interface %d missing on device %s." % (i, self.name))

        self.interfaces = collections.OrderedDict(sorted_interfaces)

    def pack_data(self) -> Optional[bytes]:
        """Pack machine data into a .tar.gz file and returns the tar content as a byte array.

        While packing files, it also applies the win2linux patch in order to remove UTF-8 BOM.

        Returns:
            bytes: the tar content.
        """
        # Make a temp folder and create a tar.gz of the lab directory
        temp_path = tempfile.mkdtemp()

        is_empty = True

        with tarfile.open("%s/hostlab.tar.gz" % temp_path, "w:gz") as tar:
            if self.folder:
                machine_files = filter(lambda x: x.is_file(), Path(self.folder).rglob("*"))

                for file in machine_files:
                    file = str(file)

                    if utils.is_excluded_file(file):
                        continue

                    # Removes the last element of the path
                    # (because it's the machine folder name and it should be included in the tar archive)
                    lab_path, machine_folder = os.path.split(self.folder)
                    (tarinfo, content) = utils.pack_file_for_tar(file,
                                                                 arc_name="hostlab/%s" % os.path.relpath(file, lab_path)
                                                                 )
                    tar.addfile(tarinfo, content)

                is_empty = False

            if self.startup_path:
                (tarinfo, content) = utils.pack_file_for_tar(self.startup_path,
                                                             arc_name="hostlab/%s.startup" % self.name
                                                             )
                tar.addfile(tarinfo, content)
                is_empty = False

            if self.shutdown_path:
                (tarinfo, content) = utils.pack_file_for_tar(self.shutdown_path,
                                                             arc_name="hostlab/%s.shutdown" % self.name
                                                             )
                tar.addfile(tarinfo, content)
                is_empty = False

            if self.lab.shared_startup_path:
                (tarinfo, content) = utils.pack_file_for_tar(self.lab.shared_startup_path,
                                                             arc_name="hostlab/shared.startup"
                                                             )
                tar.addfile(tarinfo, content)
                is_empty = False

            if self.lab.shared_shutdown_path:
                (tarinfo, content) = utils.pack_file_for_tar(self.lab.shared_shutdown_path,
                                                             arc_name="hostlab/shared.shutdown"
                                                             )
                tar.addfile(tarinfo, content)
                is_empty = False

        # If no machine files are found, return None.
        if is_empty:
            return None

        # Read tar.gz content
        with open("%s/hostlab.tar.gz" % temp_path, "rb") as tar_file:
            tar_data = tar_file.read()

        # Delete temporary tar.gz
        shutil.rmtree(temp_path)

        return tar_data

    def connect(self, terminal_name: str) -> None:
        """Connect to the device with the specified terminal.

        Args:
            terminal_name (str): The name of the terminal to use for the connection.
                The application must be correctly installed in the host system.
                This option is only visible on Linux and macOS.
                On Linux, options are /usr/bin/xterm, TMUX or an user-defined path.
                On macOS, options are Terminal (default system terminal), iTerm or TMUX.

        Returns:
            None
        """
        logging.debug("Opening terminal for device %s.", self.name)

        executable_path = utils.get_executable_path(sys.argv[0])

        if not executable_path:
            raise Exception("Unable to find Kathara.")

        is_vmachine = "-v" if self.lab.path is None else ""
        connect_command = "%s connect %s -l %s" % (executable_path, is_vmachine, self.name)
        terminal = terminal_name if terminal_name else Setting.get_instance().terminal

        logging.debug("Terminal will open in directory %s." % self.lab.path)

        def unix_connect() -> None:
            if terminal == "TMUX":
                from ..trdparty.libtmux.tmux import TMUX

                logging.debug("Attaching `%s` to TMUX session `%s` with command `%s`" % (self.name, self.lab.name,
                                                                                         connect_command))

                TMUX.get_instance().add_window(self.lab.name, self.name, connect_command, cwd=self.lab.path)
            else:
                logging.debug("Opening Linux terminal with command: %s." % connect_command)

                # Command should be passed as an array
                # https://stackoverflow.com/questions/9935151/popen-error-errno-2-no-such-file-or-directory/9935511
                subprocess.Popen([terminal, "-e", connect_command],
                                 cwd=self.lab.path,
                                 start_new_session=True
                                 )

        def windows_connect() -> None:
            complete_win_command = "& %s" % connect_command
            logging.debug("Opening Windows terminal with command: %s." % complete_win_command)
            subprocess.Popen(["powershell.exe",
                              '-Command',
                              complete_win_command
                              ],
                             creationflags=subprocess.CREATE_NEW_CONSOLE,
                             cwd=self.lab.path
                             )

        def osx_connect() -> None:
            cd_to_lab_path = "cd \"%s\" &&" % self.lab.path if self.lab.path is not None else ""
            complete_osx_command = "%s clear && %s && exit" % (cd_to_lab_path, connect_command)

            if terminal == "TMUX":
                from ..trdparty.libtmux.tmux import TMUX

                logging.debug("Attaching `%s` to TMUX session `%s` with command `%s`" % (self.name, self.lab.name,
                                                                                         complete_osx_command))

                TMUX.get_instance().add_window(self.lab.name, self.name, complete_osx_command, cwd=self.lab.path)
            else:
                import appscript
                logging.debug("Opening OSX terminal with command: %s." % complete_osx_command)
                terminal_app = appscript.app(terminal)
                if terminal == 'iTerm':
                    window = terminal_app.create_window_with_default_profile()
                    window.current_session.write(text=complete_osx_command)
                elif terminal == 'Terminal':
                    terminal_app.do_script(complete_osx_command)

        utils.exec_by_platform(unix_connect, windows_connect, osx_connect)

    def get_image(self) -> str:
        """Get the image of the device, if defined in options or machine meta. If not, use default one.

        Returns:
            str: The name of the device image.
        """
        return self.lab.general_options["image"] if "image" in self.lab.general_options else \
            self.meta["image"] if "image" in self.meta else Setting.get_instance().image

    def get_mem(self) -> str:
        """Get memory limit, if defined in options. If not, use the value from device meta. Otherwise, return None.

        Returns:
            str: The memory limit of the device.
        """
        memory = self.lab.general_options["mem"] if "mem" in self.lab.general_options else \
            self.meta["mem"] if "mem" in self.meta else None

        if memory:
            unit = memory[-1].lower()
            if unit not in ["b", "k", "m", "g"]:
                try:
                    return "%sm" % int(memory)
                except ValueError:
                    raise MachineOptionError("Memory value not valid on `%s`." % self.name)

            try:
                return "%s%s" % (int(memory[:-1]), unit)
            except ValueError:
                raise MachineOptionError("Memory value not valid on `%s`." % self.name)

        return memory

    def get_cpu(self, multiplier: int = 1) -> Optional[int]:
        """Get the CPU limit, multiplied by a specific multiplier.

        User should pass a float value ranging from 0 to max user CPUs.
        Try to took it from options, or machine meta. Otherwise, return None.

        Args:
            multiplier (int):

        Returns:
            Optional[int]: The CPU limit of the device.
        """
        if "cpus" in self.lab.general_options:
            try:
                return int(float(self.lab.general_options["cpus"]) * multiplier)
            except ValueError:
                raise MachineOptionError("CPU value not valid on `%s`." % self.name)
        elif "cpus" in self.meta:
            try:
                return int(float(self.meta["cpus"]) * multiplier)
            except ValueError:
                raise MachineOptionError("CPU value not valid on `%s`." % self.name)

        return None

    def get_ports(self) -> Optional[Dict[Tuple[int, str], int]]:
        """Get the port mapping of the device.

        Returns:
            Dict[(int, str), int]: Keys are pairs (host_port, protocol), values specifies the guest_port.
        """
        if self.meta['ports']:
            return self.meta['ports']

        return None

    def get_num_terms(self) -> int:
        """Get the number of terminal to be opened for the device.

        Returns:
            int: The number of terminal to be opened.
        """
        num_terms = 1

        if "num_terms" in self.lab.general_options:
            num_terms = self.lab.general_options['num_terms']
        elif 'num_terms' in self.meta:
            num_terms = self.meta['num_terms']

        try:
            num_terms = int(num_terms)

            if num_terms < 0:
                raise MachineOptionError("Terminals Number value on `%s` must be a positive value or zero." % self.name)
        except ValueError:
            raise MachineOptionError("Terminals Number value not valid on `%s`." % self.name)

        return num_terms

    def is_ipv6_enabled(self) -> bool:
        """Check if IPv6 is enabled on the device.

        Returns:
            bool: True if it is enabled, else False.
        """
        try:
            return bool(strtobool(self.lab.general_options["ipv6"])) if "ipv6" in self.lab.general_options else \
                bool(strtobool(self.meta["ipv6"])) if "ipv6" in self.meta else Setting.get_instance().enable_ipv6
        except ValueError:
            raise MachineOptionError("IPv6 value not valid on `%s`." % self.name)

    def update_meta(self, args: Dict[str, Any]) -> None:
        """Update the device metas from a dict.

        Args:
            args (Dict[str, Any]): Keys are the meta properties names, values are the updated meta properties values.

        Returns:
            None
        """
        if 'exec_commands' in args and args['exec_commands'] is not None:
            for command in args['exec_commands']:
                self.add_meta("exec", command)

        if 'mem' in args and args['mem'] is not None:
            self.add_meta("mem", args['mem'])

        if 'cpus' in args and args['cpus'] is not None:
            self.add_meta("cpus", args['cpus'])

        if 'image' in args and args['image'] is not None:
            self.add_meta("image", args['image'])

        if 'bridged' in args and args['bridged'] is not None and args['bridged']:
            self.add_meta("bridged", True)

        if 'ports' in args and args['ports'] is not None:
            for port in args['ports']:
                self.add_meta("port", port)

        if 'num_terms' in args and args['num_terms'] is not None:
            self.add_meta('num_terms', args['num_terms'])

        if 'sysctls' in args and args['sysctls'] is not None:
            for sysctl in args['sysctls']:
                self.add_meta("sysctl", sysctl)

    def __repr__(self) -> str:
        return "Machine(%s, %s, %s)" % (self.name, self.interfaces, self.meta)



from ..console_menu import MenuItem
from ..console_menu import ExitItem
from .external_item import ExternalItem
from .command_item import CommandItem
from .function_item import FunctionItem
from .submenu_item import SubmenuItem
from .selection_item import SelectionItem

__all__ = ['CommandItem', 'ExitItem', 'ExternalItem', 'FunctionItem', 'MenuItem', 'SelectionItem', 'SubmenuItem']



import argparse
import logging
import re
import sys
from typing import List

from ... import utils
from ...foundation.cli.command.Command import Command
from ...manager.Kathara import Kathara
from ...model.Lab import Lab
from ...strings import strings, wiki_description


class LconfigCommand(Command):
    def __init__(self) -> None:
        Command.__init__(self)

        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(
            prog='kathara lconfig',
            description=strings['lconfig'],
            epilog=wiki_description,
            add_help=False
        )

        self.parser.add_argument(
            '-h', '--help',
            action='help',
            default=argparse.SUPPRESS,
            help='Show an help message and exit.'
        )
        self.parser.add_argument(
            '-d', '--directory',
            metavar='LAB_PATH',
            required=False,
            help='Path of the lab to configure, if not specified the current path is used'
        )
        self.parser.add_argument(
            '-n', '--name',
            metavar='DEVICE_NAME',
            required=True,
            help='Name of the device to be connected on desired collision domains.'
        )
        self.parser.add_argument(
            '--eth',
            dest='eths',
            metavar='CD',
            nargs='+',
            required=True,
            help='Specify the collision domain for an interface.'
        )

    def run(self, current_path: str, argv: List[str]) -> None:
        self.parse_args(argv)
        args = self.get_args()

        lab_path = args['directory'].replace('"', '').replace("'", '') if args['directory'] else current_path
        lab_path = utils.get_absolute_path(lab_path)

        for eth in args['eths']:
            # Only alphanumeric characters are allowed
            matches = re.search(r"^\w+$", eth)

            if not matches:
                sys.stderr.write('Syntax error in --eth field.\n')
                self.parser.print_help()
                exit(1)

        lab = Lab(None, path=lab_path)

        device = lab.get_or_new_machine(args['name'])
        device.api_object = Kathara.get_instance().get_machine_api_object(lab.hash, args['name'])

        for eth in args['eths']:
            logging.info("Adding interface to device `%s` for collision domain `%s`..." % (args['name'], eth))
            lab.connect_machine_to_link(args['name'], eth)

        Kathara.get_instance().update_lab(lab)



import sys
from abc import ABC, abstractmethod
from typing import Any, Optional, Callable

import pyuv


class Terminal(ABC):
    __slots__ = ['handler', '_loop', '_system_stdin', '_system_stdout', '_external_terminal', '_resize_signal']

    def __init__(self, handler: Any) -> None:
        self.handler: Any = handler

        self._loop: Optional[pyuv.Loop] = None
        self._system_stdin: Optional[pyuv.TTY] = None
        self._system_stdout: Optional[pyuv.TTY] = None
        self._external_terminal: Optional[pyuv.TTY] = None
        self._resize_signal: Optional[pyuv.Signal] = None

    def start(self) -> None:
        self._loop = pyuv.Loop.default_loop()

        self._system_stdin = pyuv.TTY(self._loop, sys.stdin.fileno(), True)
        self._system_stdin.set_mode(1)
        self._system_stdin.start_read(self._write_on_external_terminal())

        self._system_stdout = pyuv.TTY(self._loop, sys.stdout.fileno(), False)

        self._start_external()

        self._resize_terminal()

        self._loop.run()

        pyuv.TTY.reset_mode()

    @abstractmethod
    def _start_external(self) -> None:
        pass

    def close(self) -> None:
        self._on_close()

        self._system_stdin.close()
        self._system_stdout.close()
        self._external_terminal.close()

        self._loop.stop()

    @abstractmethod
    def _on_close(self) -> None:
        pass

    def _write_on_external_terminal(self) -> Callable:
        def write_on_external_terminal(handle, data, error):
            self._external_terminal.write(data)

        return write_on_external_terminal

    def _read_external_terminal(self) -> Callable:
        def read_external_terminal(handle, data, error):
            if data:
                self._system_stdout.write(data)

                if data == b"\r\nexit\r\n":
                    self.close()
            else:
                self.close()

        return read_external_terminal

    @abstractmethod
    def _resize_terminal(self) -> None:
        pass



import argparse
import logging
import os
import sys
from typing import List

from ... import utils
from ...exceptions import PrivilegeError
from ...foundation.cli.command.Command import Command
from ...manager.Kathara import Kathara
from ...model.Lab import Lab
from ...parser.netkit.DepParser import DepParser
from ...parser.netkit.ExtParser import ExtParser
from ...parser.netkit.FolderParser import FolderParser
from ...parser.netkit.LabParser import LabParser
from ...parser.netkit.OptionParser import OptionParser
from ...setting.Setting import Setting
from ...strings import strings, wiki_description


class LstartCommand(Command):
    def __init__(self) -> None:
        Command.__init__(self)

        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(
            prog='kathara lstart',
            description=strings['lstart'],
            epilog=wiki_description,
            add_help=False
        )

        self.parser.add_argument(
            '-h', '--help',
            action='help',
            default=argparse.SUPPRESS,
            help='Show an help message and exit.'
        )

        group = self.parser.add_mutually_exclusive_group(required=False)

        group.add_argument(
            "--noterminals",
            action="store_const",
            dest="terminals",
            const=False,
            default=None,
            help='Start the lab without opening terminal windows.'
        )
        group.add_argument(
            "--terminals",
            action="store_const",
            dest="terminals",
            const=True,
            help='Start the lab opening terminal windows.'
        )
        group.add_argument(
            "--privileged",
            action="store_const",
            const=True,
            required=False,
            help='Start the devices in privileged mode. MUST BE ROOT FOR THIS OPTION.'
        )
        self.parser.add_argument(
            '-d', '--directory',
            required=False,
            help='Specify the folder containing the lab.'
        )
        self.parser.add_argument(
            '-F', '--force-lab',
            dest='force_lab',
            required=False,
            action='store_true',
            help='Force the lab to start without a lab.conf or lab.dep file.'
        )
        self.parser.add_argument(
            '-l', '--list',
            required=False,
            action='store_true',
            help='Show information about running devices after the lab has been started.'
        )
        self.parser.add_argument(
            '-o', '--pass',
            dest='options',
            metavar="OPTION",
            nargs='*',
            required=False,
            help="Apply options to all devices of a lab during startup."
        )
        self.parser.add_argument(
            '--xterm',
            required=False,
            help='Set a different terminal emulator application (Unix only).'
        )
        self.parser.add_argument(
            '--print',
            dest="dry_mode",
            required=False,
            action='store_true',
            help='Open the lab.conf file and check if it is correct (dry run).'
        )
        self.parser.add_argument(
            '-H', '--no-hosthome',
            dest="no_hosthome",
            action="store_const",
            const=False,
            help='/hosthome dir will not be mounted inside the devices.'
        )
        self.parser.add_argument(
            '-S', '--no-shared',
            dest="no_shared",
            action="store_const",
            const=False,
            help='/shared dir will not be mounted inside the devices.'
        )
        self.parser.add_argument(
            'machine_name',
            metavar='DEVICE_NAME',
            nargs='*',
            help='Launches only specified devices.'
        )

    def run(self, current_path: str, argv: List[str]) -> Lab:
        self.parse_args(argv)
        args = self.get_args()

        lab_path = args['directory'].replace('"', '').replace("'", '') if args['directory'] else current_path
        lab_path = utils.get_absolute_path(lab_path)

        Setting.get_instance().open_terminals = args['terminals'] if args['terminals'] is not None \
            else Setting.get_instance().open_terminals
        Setting.get_instance().terminal = args['xterm'] or Setting.get_instance().terminal

        if args['privileged']:
            if not utils.is_admin():
                raise Exception("You must be root in order to start Kathara devices in privileged mode.")
            else:
                logging.warning("Running devices with privileged capabilities, terminals won't open!")
                Setting.get_instance().open_terminals = False

        if args['dry_mode']:
            logging.info(utils.format_headers("Checking Lab"))
        else:
            logging.info(utils.format_headers("Starting Lab"))

        try:
            lab = LabParser.parse(lab_path)
        except IOError as e:
            if not args['force_lab']:
                raise Exception(str(e))
            else:
                lab = FolderParser.parse(lab_path)

        # Reorder machines by lab.dep file, if present.
        dependencies = DepParser.parse(lab_path)
        if dependencies:
            lab.apply_dependencies(dependencies)

        lab_meta_information = str(lab)

        if lab_meta_information:
            logging.info("\n" + lab_meta_information)
            logging.info(utils.format_headers())

        if len(lab.machines) <= 0:
            raise Exception("No devices in the current lab. Exiting...")

        try:
            options = OptionParser.parse(args['options'])
            lab.general_options = {**lab.general_options, **options}
        except ValueError as e:
            raise e

        lab_ext_path = os.path.join(lab_path, 'lab.ext')
        if os.path.exists(lab_ext_path):
            if utils.is_platform(utils.LINUX) or utils.is_platform(utils.LINUX2):
                if utils.is_admin():
                    external_links = ExtParser.parse(lab_path)

                    if external_links:
                        lab.attach_external_links(external_links)

                        # Since xterm does not work with "sudo", we do not open terminals when lab.ext is present.
                        Setting.get_instance().open_terminals = False
                else:
                    raise PrivilegeError("You must be root in order to use lab.ext file.")
            else:
                raise OSError("lab.ext is only available on UNIX systems.")

        # If dry mode, we just check if the lab.conf is correct.
        if args['dry_mode']:
            logging.info("lab.conf file is correct.")
            if dependencies:
                logging.info("lab.dep file is correct.")
            if os.path.exists(lab_ext_path):
                logging.info("lab.ext file is correct.")

            logging.info("Exiting...")

            sys.exit(0)

        lab.add_option('hosthome_mount', args['no_hosthome'])
        lab.add_option('shared_mount', args['no_shared'])
        lab.add_option('privileged_machines', args['privileged'])

        Kathara.get_instance().deploy_lab(lab, selected_machines=args['machine_name'])

        if args['list']:
            print(next(Kathara.get_instance().get_formatted_lab_info(lab.hash)))

        return lab



import io
from abc import ABC, abstractmethod
from typing import Dict, Set, Any, Generator, Tuple, List

from ...model.Lab import Lab
from ...model.Machine import Machine


class IManager(ABC):
    @abstractmethod
    def deploy_lab(self, lab: Lab, selected_machines: Set[str] = None) -> None:
        """
        Deploy a Kathara network scenario.

        Args:
            lab (Kathara.model.Lab): A Kathara network scenario.
            selected_machines (Set[str]): If not None, deploy only the specified devices.

        Returns:
            None
        """
        raise NotImplementedError("You must implement `deploy_lab` method.")

    @abstractmethod
    def update_lab(self, lab: Lab) -> None:
        """
        Update a running network scenario.

        Args:
            lab (Kathara.model.Lab): A Kathara network scenario.

        Returns:
            None
        """
        raise NotImplementedError("You must implement `update_lab` method.")

    @abstractmethod
    def undeploy_lab(self, lab_hash: str, selected_machines: Set[str] = None) -> None:
        """
        Undeploy a Kathara network scenario.

        Args:
            lab_hash (str): The hash of the network scenario to undeploy.
            selected_machines (Set[str]): If not None, undeploy only the specified devices.

        Returns:
            None
        """
        raise NotImplementedError("You must implement `undeploy_lab` method.")

    @abstractmethod
    def wipe(self, all_users: bool = False) -> None:
        """
        Undeploy all the running network scenarios.

        Args:
        all_users (bool): If false, undeploy only the current user network scenarios. If true, undeploy the
           running network scenarios of all users.

        Returns:
            None
        """
        raise NotImplementedError("You must implement `wipe` method.")

    @abstractmethod
    def connect_tty(self, lab_hash: str, machine_name: str, shell: str = None, logs: bool = False) -> None:
        """
        Connect to a device in a running network scenario, using the specified shell.

        Args:
            lab_hash (str): The hash of the network scenario to undeploy.
            machine_name (str): The name of the device to connect.
            shell (str): The name of the shell to use for connecting.
            logs (bool): If True, print startup logs on stdout.

        Returns:
            None
        """
        raise NotImplementedError("You must implement `connect_tty` method.")

    @abstractmethod
    def exec(self, lab_hash: str, machine_name: str, command: str) -> Generator[Tuple[bytes, bytes], None, None]:
        """
        Exec a command on a device in a running network scenario.
        Args:
            lab_hash (str): The hash of the network scenario to undeploy.
            machine_name (str): The name of the device to connect.
            command (str): The command to exec on the device.

        Returns:
            Generator[Tuple[bytes, bytes]]: A generator of tuples containing the stdout and stderr in bytes.
        """
        raise NotImplementedError("You must implement `exec` method.")

    @abstractmethod
    def copy_files(self, machine: Machine, guest_to_host: Dict[str, io.IOBase]) -> None:
        """
        Copy files on a running device in the specified paths.

        Args:
            machine (Kathara.model.Machine): A running machine object. It must have the api_object field populated.
            guest_to_host (Dict[str, io.IOBase]): A dict containing the device path as key and
                fileobj to copy in path as value.

        Returns:
            None
        """
        raise NotImplementedError("You must implement `copy_files` method.")

    @abstractmethod
    def get_lab_info(self, lab_hash: str = None, machine_name: str = None, all_users: bool = False) -> \
            Generator[Dict[str, Any], None, None]:
        """
        Return information about the running devices.

        Args:
            lab_hash (str): If not None, return information of the corresponding network scenario.
            machine_name (str): If not None, return information of the specified device.
            all_users (bool): If True, return information about the device of all users.

        Returns:
              Generator[Dict[str, Any], None, None]: A generator containing dicts containing device names as keys and
              their info as values.
        """
        raise NotImplementedError("You must implement `get_lab_info` method.")

    @abstractmethod
    def get_formatted_lab_info(self, lab_hash: str = None, machine_name: str = None, all_users: bool = False) -> str:
        """
        Return a formatted string with the information about the running devices.

        Args:
            lab_hash (str): If not None, return information of the corresponding network scenario.
            machine_name (str): If not None, return information of the specified device.
            all_users (bool): If True, return information about the device of all users.

        Returns:
             str: String containing devices info
        """
        raise NotImplementedError("You must implement `get_formatted_lab_info` method.")

    @abstractmethod
    def get_machine_api_object(self, lab_hash: str, machine_name: str) -> Any:
        """
        Return the corresponding API object of a running device in a network scenario.

        Args:
            lab_hash (str): The hash of the network scenario.
            machine_name (str): The name of the device.

        Returns:
            Any: The Api object of the device specific for the current manager.
        """
        raise NotImplementedError("You must implement `get_machine_api_object` method.")

    @abstractmethod
    def get_machine_info(self, machine_name: str, lab_hash: str = None, all_users: bool = False) \
            -> List[Dict[str, Any]]:
        """
        Return information of running devices with a specified name.

        Args:
            machine_name (str): The device name.
            lab_hash (str): If not None, search the device in the specified network scenario.
            all_users (bool): If True, search the device among all the users devices.

        Returns:
            List[Dict[str, Any]]: A list of dicts containing the devices info.
        """
        raise NotImplementedError("You must implement `get_machine_info` method.")

    @abstractmethod
    def get_formatted_machine_info(self, machine_name: str, lab_hash: str = None, all_users: bool = False) -> str:
        """
        Return formatted information of running devices with a specified name.

        Args:
            machine_name (str): The device name.
            lab_hash (str): If not None, search the device in the specified network scenario.
            all_users (bool): If True, search the device among all the users devices.

        Returns:
            str: The formatted devices properties.
        """
        raise NotImplementedError("You must implement `get_formatted_machine_info` method.")

    @abstractmethod
    def check_image(self, image_name: str) -> None:
        """
        Check if the specified image is valid.

        Args:
            image_name (str): The name of the image

        Returns:
            None

        Raises:
            ConnectionError: The image is not locally available and there is no connection to a remote image repository.
            Exception: The image is not found.
        """
        raise NotImplementedError("You must implement `check_image` method.")

    @abstractmethod
    def get_release_version(self) -> str:
        """
        Return the current manager version.

        Returns:
            str: The current manager version.
        """
        raise NotImplementedError("You must implement `get_release_version` method.")

    @staticmethod
    def get_formatted_manager_name() -> str:
        """
        Return a formatted string containing the current manager name.

        Returns:
            str: A formatted string containing the current manager name.
        """
        raise NotImplementedError("You must implement `get_formatted_manager_name` method.")



import argparse
import logging
import re
import sys
from typing import List

from ...foundation.cli.command.Command import Command
from ...manager.Kathara import Kathara
from ...model.Lab import Lab
from ...strings import strings, wiki_description


class VconfigCommand(Command):
    def __init__(self) -> None:
        Command.__init__(self)

        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(
            prog='kathara vconfig',
            description=strings['vconfig'],
            epilog=wiki_description,
            add_help=False
        )

        self.parser.add_argument(
            '-h', '--help',
            action='help',
            default=argparse.SUPPRESS,
            help='Show an help message and exit.'
        )
        self.parser.add_argument(
            '-n', '--name',
            metavar='DEVICE_NAME',
            required=True,
            help='Name of the device to be connected on desired collision domains.'
        )
        self.parser.add_argument(
            '--eth',
            dest='eths',
            metavar='CD',
            nargs='+',
            required=True,
            help='Specify the collision domain for an interface.'
        )

    def run(self, current_path: str, argv: List[str]) -> None:
        self.parse_args(argv)
        args = self.get_args()

        for eth in args['eths']:
            # Only alphanumeric characters are allowed
            matches = re.search(r"^\w+$", eth)

            if not matches:
                sys.stderr.write('Syntax error in --eth field.\n')
                self.parser.print_help()
                exit(1)

        lab = Lab("kathara_vlab")

        device = lab.get_or_new_machine(args['name'])
        device.api_object = Kathara.get_instance().get_machine_api_object(lab.hash, args['name'])

        for eth in args['eths']:
            logging.info("Adding interface to device `%s` for collision domain `%s`..." % (args['name'], eth))
            lab.connect_machine_to_link(args['name'], eth)

        Kathara.get_instance().update_lab(lab)



import signal
from typing import Any

import pyuv
from docker import DockerClient

from ....foundation.manager.terminal.Terminal import Terminal


class DockerTTYTerminal(Terminal):
    __slots__ = ['client', 'exec_id']

    def __init__(self, handler: Any, client: DockerClient, exec_id: str):
        super().__init__(handler)

        self.client: DockerClient = client
        self.exec_id: str = exec_id

    def _start_external(self) -> None:
        self._external_terminal = pyuv.TTY(self._loop, self.handler.fileno(), True)
        self._external_terminal.start_read(self._read_external_terminal())

    def _on_close(self) -> None:
        self._system_stdin.set_mode(0)

        self._resize_signal.close()

    def _resize_terminal(self) -> None:
        def resize_terminal(signal_handle, signal_num):
            w, h = self._system_stdin.get_winsize()
            self.client.api.exec_resize(self.exec_id, height=h, width=w)

        self._resize_signal = pyuv.Signal(self._loop)
        self._resize_signal.start(resize_terminal, signal.SIGWINCH)

        # Run first time to set the proper terminal size
        resize_terminal(None, None)



import base64
import hashlib
import importlib
import io
import logging
import math
import os
import re
import shutil
import sys
import tarfile
import tempfile
from io import BytesIO
from itertools import islice
from multiprocessing import cpu_count
from platform import node, machine
from sys import platform as _platform
from typing import Any, Optional, Match, Generator, List, Callable, Union, Dict, Iterable

from binaryornot.check import is_binary
from slug import slug

from .trdparty.consolemenu import PromptUtils, Screen

# Platforms constants definition.
MAC_OS: str = "darwin"
WINDOWS: str = "win32"
LINUX: str = "linux"
LINUX2: str = "linux2"

# List of ignored files
EXCLUDED_FILES: List[str] = ['.DS_Store']

# True if launched from CLI
CLI_ENV: bool = False

# Reserved names for devices
RESERVED_MACHINE_NAMES: List[str] = ['shared', '_test']


# Generic Functions
def check_python_version() -> None:
    if sys.version_info < (3, 0):
        logging.critical("Python version should be greater than 3.0")
        sys.exit(1)


def class_for_name(module_name: str, class_name: str) -> Any:
    m = importlib.import_module(module_name + "." + class_name)
    return getattr(m, class_name)


def generate_urlsafe_hash(string: str) -> str:
    string = re.sub(r'[^\x00-\x7F]+', '', string)
    return base64.urlsafe_b64encode(hashlib.md5(string.encode('utf-8', errors='ignore')).digest())[:-2] \
        .decode('utf-8') \
        .replace('-', '').replace('_', '')


def get_absolute_path(path: str) -> str:
    abs_path = os.path.realpath(path)
    return abs_path if not os.path.islink(abs_path) else os.readlink(abs_path)


def get_executable_path(exec_path: str) -> Optional[str]:
    exec_abs_path = get_absolute_path(exec_path)

    if os.path.exists(exec_abs_path) and os.path.isfile(exec_abs_path):
        # If kathara is launched as a python script
        exec_abs_path = "\"" + exec_abs_path + "\""
        if exec_path.endswith(".py"):
            # Prepend python in windows because it has no shebang
            return exec_by_platform(lambda: exec_abs_path,
                                    lambda: "%s %s" % (sys.executable, exec_abs_path),
                                    lambda: "%s %s" % (sys.executable, exec_abs_path)
                                    )
        else:
            # Maybe the executable is not in path, but is still a binary file
            return exec_abs_path
    else:
        # If kathara is in the path, return the absolute path of kathara
        which_path = shutil.which(exec_path)
        if which_path:
            return "\"" + which_path + "\""

    return None


def re_search_fail(expression: str, line: str) -> Match:
    matches = re.search(expression, line)

    if not matches:
        raise ValueError()

    return matches


def list_chunks(iterable: List, size: int) -> Generator:
    it = iter(iterable)
    item = list(islice(it, size))
    while item:
        yield item
        item = list(islice(it, size))


def chunk_list(iterable: Union[List, Iterable], size: int) -> List[List]:
    return [iterable] if len(iterable) < size else list_chunks(iterable, size)


def confirmation_prompt(prompt_string: str, callback_yes: Callable, callback_no: Callable) -> Any:
    prompt_utils = PromptUtils(Screen())
    answer = prompt_utils.prompt_for_bilateral_choice(prompt_string, 'y', 'n')

    if answer == "n":
        return callback_no()

    return callback_yes()


def get_pool_size() -> int:
    return cpu_count()


# Platform Specific Functions
def is_platform(desired_platform: str) -> bool:
    return _platform == desired_platform


def exec_by_platform(fun_linux: Callable, fun_windows: Callable, fun_mac: Callable) -> Any:
    if _platform == LINUX or _platform == LINUX2:
        return fun_linux()
    elif _platform == WINDOWS:
        return fun_windows()
    elif _platform == MAC_OS:
        return fun_mac()


# Architecture Test
def get_architecture() -> str:
    architecture = machine().lower()

    logging.debug("Machine architecture is `%s`." % architecture)

    if architecture == "x86_64" or architecture == "amd64":
        return "amd64"
    elif architecture == "i686":
        return "386"
    elif architecture in ['arm64', 'aarch64']:
        return "arm64"
    elif architecture == "armv7l":
        return "armv7"
    elif architecture == "armv6l":
        return "armv6"
    else:
        raise Exception("Not implemented for %s." % architecture)


def convert_win_2_linux(filename: str) -> bytes:
    if not is_binary(filename):
        file_content = open(filename, mode='r', encoding='utf-8-sig').read()
        return file_content.replace("\n\r", "\n").encode('utf-8')

    return open(filename, mode='rb').read()


def is_admin() -> bool:
    def unix_root():
        return os.getuid() == 0

    def windows_admin():
        import ctypes
        return ctypes.windll.shell32.IsUserAnAdmin() != 0

    return exec_by_platform(unix_root, windows_admin, unix_root)


def get_current_user_home() -> str:
    def passwd_home():
        user_info = get_current_user_info()
        return user_info.pw_dir

    def default_home():
        return os.path.expanduser('~')

    return exec_by_platform(passwd_home, default_home, default_home)


def get_current_user_uid_gid() -> (int, int):
    def unix():
        user_info = get_current_user_info()
        return user_info.pw_uid, user_info.pw_gid

    return exec_by_platform(unix, lambda: (None, None), unix)


def get_current_user_name() -> str:
    hostname = generate_urlsafe_hash(node())

    def unix():
        user_info = get_current_user_info()
        return user_info.pw_name

    def windows():
        import getpass
        return getpass.getuser()

    return slug("%s-%s" % (exec_by_platform(unix, windows, unix), hostname))


def get_current_user_info() -> Any:
    def passwd_info():
        import pwd
        user_id = os.getuid()

        # It's root, take the real user from env variable
        if user_id == 0:
            # If it's a sudoer, take the SUDO_UID and use it as user_id variable
            # If not, keep using the user_id = 0
            real_user_id = os.environ.get("SUDO_UID")

            if real_user_id:
                user_id = int(real_user_id)

        return pwd.getpwuid(user_id)

    return exec_by_platform(passwd_info, lambda: None, passwd_info)


# Formatting Functions
def format_headers(message: str = "") -> str:
    footer = "=============================="
    half_message = int((len(message) / 2) + 1)
    second_half_message = half_message

    if len(message) % 2 == 0:
        second_half_message -= 1

    message = " " + message + " " if message != "" else "=="
    return footer[half_message:] + message + footer[second_half_message:]


def human_readable_bytes(size_bytes: int) -> str:
    if size_bytes == 0:
        return "0 B"

    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)

    return "%s %s" % (s, size_name[i])


# Lab Functions
def get_lab_temp_path(lab_name: str, force_creation: bool = True) -> str:
    def windows_path():
        import win32file
        return win32file.GetLongPathName(tempfile.gettempdir())

    tempdir = exec_by_platform(tempfile.gettempdir,
                               windows_path,
                               lambda: re.sub(r"/+", "/", "/%s" % get_absolute_path("/tmp"))
                               )
    lab_temp_directory = os.path.join(tempdir, lab_name)
    if not os.path.isdir(lab_temp_directory) and force_creation:
        os.mkdir(lab_temp_directory)

    return lab_temp_directory


def get_vlab_temp_path(force_creation: bool = True) -> str:
    return get_lab_temp_path("kathara_vlab", force_creation=force_creation)


def pack_file_for_tar(file_obj: Union[str, io.IOBase], arc_name: str) -> (tarfile.TarInfo, bytes):
    if isinstance(file_obj, str):
        file_content_patched = convert_win_2_linux(file_obj)
        file_content = BytesIO(file_content_patched)
        filesize = len(file_content_patched)
    elif isinstance(file_obj, io.IOBase):
        file_content = file_obj.read()
        file_content = BytesIO(file_content.encode('utf-8') if isinstance(file_obj, io.TextIOBase) else file_content)
        file_obj.seek(0, 2)
        filesize = file_obj.tell()
        file_obj.seek(0)
    else:
        raise ValueError("File type %s not supported" % type(file_obj))

    tarinfo = tarfile.TarInfo(arc_name.replace("\\", "/"))  # Tar files must have Linux-style paths
    tarinfo.size = filesize

    return tarinfo, file_content


def pack_files_for_tar(guest_to_host: Dict) -> bytes:
    with tempfile.NamedTemporaryFile(mode='wb+', suffix='.tar.gz') as temp_file:
        with tarfile.open(fileobj=temp_file, mode='w:gz') as tar_file:
            for path, file_obj in guest_to_host.items():
                tar_info, file_content = pack_file_for_tar(file_obj, arc_name=path)
                tar_file.addfile(tar_info, file_content)

        temp_file.seek(0)

        tar_data = temp_file.read()

    return tar_data


def is_excluded_file(path: str) -> bool:
    _, filename = os.path.split(path)

    return filename in EXCLUDED_FILES



import logging
import mmap
import os
import re
from typing import List, Optional

from ...trdparty.depgen import depgen


class DepParser(object):
    """
    Class responsible for parsing the lab.dep file.
    """
    @staticmethod
    def parse(path: str) -> Optional[List[str]]:
        """
        Parse the lab.dep file and return a List of string containing the names of the device ordered considering the
        dependencies.

        Args:
            path (str): The path to the lab.dep file.

        Returns:
            Optional[List[str]]: A List of string containing the names of the device ordered considering the
            dependencies.
        """
        lab_dep_path = os.path.join(path, 'lab.dep')

        if not os.path.exists(lab_dep_path):
            return None

        if os.stat(lab_dep_path).st_size == 0:
            logging.warning("lab.dep file is empty. Ignoring...")
            return None

        dependencies = {}

        # Reads lab.dep in memory so it is faster.
        try:
            with open(lab_dep_path, 'r') as dep_file:
                dep_mem_file = mmap.mmap(dep_file.fileno(), 0, access=mmap.ACCESS_READ)
        except Exception:
            raise IOError("Cannot open lab.dep file.")

        line_number = 1
        line = dep_mem_file.readline().decode('utf-8')
        while line:
            line = line.strip()
            if not line or line.startswith('#'):
                continue
            # E.g. MACHINE: MACHINE1 MACHINE2 MACHINE3
            # Or MACHINE:MACHINE1 MACHINE2 MACHINE3
            matches = re.search(r"^(?P<key>\w+):\s?(?P<deps>(\w+ ?)+)$",
                                line
                                )

            if matches:
                key = matches.group("key").strip()
                deps = matches.group("deps").strip().split(" ")

                # Dependencies are saved as dependencies[machine3] = [machine1, machine2]
                dependencies[key] = deps
            else:
                raise Exception("[ERROR] In lab.dep - line %d: Syntax error." % line_number)

            line_number += 1
            line = dep_mem_file.readline().decode('utf-8')

        if depgen.has_loop(dependencies):
            raise Exception("ERROR: Loop in lab.dep.\n")

        return depgen.flatten(dependencies)



from __future__ import annotations

import io
from typing import Set, Dict, Generator, Any, Tuple, List

from ..foundation.manager.IManager import IManager
from ..foundation.manager.ManagerFactory import ManagerFactory
from ..model.Lab import Lab
from ..model.Machine import Machine
from ..setting.Setting import Setting, AVAILABLE_MANAGERS


class Kathara(IManager):
    """Facade class for interacting with Kathara. It is a proxy for the selected virtualization manager."""
    __slots__ = ['manager']

    __instance: Kathara = None

    @staticmethod
    def get_instance() -> Kathara:
        """Get an instance of the current Kathara Manager.
    
        Returns:
            Kathara: instance of the current Kathara Manager.
        """
        if Kathara.__instance is None:
            Kathara()

        return Kathara.__instance

    def __init__(self) -> None:
        if Kathara.__instance is not None:
            raise Exception("This class is a singleton!")
        else:
            manager_type = Setting.get_instance().manager_type

            self.manager: IManager = ManagerFactory().create_instance(module_args=(manager_type,),
                                                                      class_args=(manager_type.capitalize(),)
                                                                      )

            Kathara.__instance = self

    def deploy_lab(self, lab: Lab, selected_machines: Set[str] = None) -> None:
        """Deploy a Kathara network scenario.

        Args:
            lab (Kathara.model.Lab): A Kathara network scenario.
            selected_machines (Set[str]): If not None, deploy only the specified devices.

        Returns:
            None
        """
        self.manager.deploy_lab(lab, selected_machines)

    def update_lab(self, lab: Lab) -> None:
        """Update a running network scenario.

        Args:
            lab (Kathara.model.Lab): A Kathara network scenario.

        Returns:
            None
        """
        self.manager.update_lab(lab)

    def undeploy_lab(self, lab_hash: str, selected_machines: Set[str] = None) -> None:
        """Undeploy a Kathara network scenario.

        Args:
            lab_hash (str): The hash of the network scenario to undeploy.
            selected_machines (Set[str]): If not None, undeploy only the specified devices.

        Returns:
            None
        """
        self.manager.undeploy_lab(lab_hash, selected_machines)

    def wipe(self, all_users: bool = False) -> None:
        """Undeploy all the running network scenarios.

        Args:
        all_users (bool): If false, undeploy only the current user network scenarios. If true, undeploy the
           running network scenarios of all users.

        Returns:
            None
        """
        self.manager.wipe(all_users)

    def connect_tty(self, lab_hash: str, machine_name: str, shell: str = None, logs: bool = False) -> None:
        """Connect to a device in a running network scenario, using the specified shell.

        Args:
            lab_hash (str): The hash of the network scenario to undeploy.
            machine_name (str): The name of the device to connect.
            shell (str): The name of the shell to use for connecting.
            logs (bool): If True, print startup logs on stdout.

        Returns:
            None
        """
        self.manager.connect_tty(lab_hash, machine_name, shell, logs)

    def exec(self, lab_hash: str, machine_name: str, command: str) -> Generator[Tuple[bytes, bytes], None, None]:
        """Exec a command on a device in a running network scenario.

        Args:
            lab_hash (str): The hash of the network scenario to undeploy.
            machine_name (str): The name of the device to connect.
            command (str): The command to exec on the device.

        Returns:
            Generator[Tuple[bytes, bytes]]: A generator of tuples containing the stdout and stderr in bytes.
        """
        return self.manager.exec(lab_hash, machine_name, command)

    def copy_files(self, machine: Machine, guest_to_host: Dict[str, io.IOBase]) -> None:
        """Copy files on a running device in the specified paths.

        Args:
            machine (Kathara.model.Machine): A running machine object. It must have the api_object field populated.
            guest_to_host (Dict[str, io.IOBase]): A dict containing the device path as key and
                fileobj to copy in path as value.

        Returns:
            None
        """
        self.manager.copy_files(machine, guest_to_host)

    def get_lab_info(self, lab_hash: str = None, machine_name: str = None, all_users: bool = False) -> \
            Generator[Dict[str, Any], None, None]:
        """Return information about the running devices.

        Args:
            lab_hash (str): If not None, return information of the corresponding network scenario.
            machine_name (str): If not None, return information of the specified device.
            all_users (bool): If True, return information about the device of all users.

        Returns:
              Generator[Dict[str, Any], None, None]: A generator containing dicts containing device names as keys and
              their info as values.
        """
        return self.manager.get_lab_info(lab_hash, machine_name, all_users)

    def get_formatted_lab_info(self, lab_hash: str = None, machine_name: str = None, all_users: bool = False) -> str:
        """Return a formatted string with the information about the running devices.

        Args:
            lab_hash (str): If not None, return information of the corresponding network scenario.
            machine_name (str): If not None, return information of the specified device.
            all_users (bool): If True, return information about the device of all users.

        Returns:
             str: String containing devices info
        """
        return self.manager.get_formatted_lab_info(lab_hash, machine_name, all_users)

    def get_machine_api_object(self, lab_hash: str, machine_name: str) -> Any:
        """Return the corresponding API object of a running device in a network scenario.

        Args:
            lab_hash (str): The hash of the network scenario.
            machine_name (str): The name of the device.

        Returns:
            Any: The Api object of the device specific for the current manager.
        """
        return self.manager.get_machine_api_object(lab_hash, machine_name)

    def get_machine_info(self, machine_name: str, lab_hash: str = None, all_users: bool = False) \
            -> List[Dict[str, Any]]:
        """Return information of running devices with a specified name.

        Args:
            machine_name (str): The device name.
            lab_hash (str): If not None, search the device in the specified network scenario.
            all_users (bool): If True, search the device among all the users devices.

        Returns:
            List[Dict[str, Any]]: A list of dicts containing the devices info.
        """
        return self.manager.get_machine_info(machine_name, lab_hash, all_users)

    def get_formatted_machine_info(self, machine_name: str, lab_hash: str = None, all_users: bool = False) -> str:
        """Return formatted information of running devices with a specified name.

        Args:
            machine_name (str): The device name.
            lab_hash (str): If not None, search the device in the specified network scenario.
            all_users (bool): If True, search the device among all the users devices.

        Returns:
            str: The formatted devices properties.
        """
        return self.manager.get_formatted_machine_info(machine_name, lab_hash, all_users)

    def check_image(self, image_name: str) -> None:
        """Check if the specified image is valid.

        Args:
            image_name (str): The name of the image

        Returns:
            None

        Raises:
            ConnectionError: The image is not locally available and there is no connection to a remote image repository.
            Exception: The image is not found.
        """
        self.manager.check_image(image_name)

    def get_release_version(self) -> str:
        """Return the current manager version.

        Returns:
            str: The current manager version.
        """
        return self.manager.get_release_version()

    def get_formatted_manager_name(self) -> str:
        """Return a formatted string containing the current manager name.

        Returns:
            str: A formatted string containing the current manager name.
        """
        return self.manager.get_formatted_manager_name()

    @staticmethod
    def get_available_managers_name() -> Dict[str, str]:
        """Return a dict containing the available manager names.

        Returns:
            Dict[str, str]: keys are the manager names, values are the formatted manager names.
        """
        managers = {}
        manager_factory = ManagerFactory()

        for manager_name in AVAILABLE_MANAGERS:
            manager = manager_factory.get_class(module_args=(manager_name,),
                                                class_args=(manager_name.capitalize(),)
                                                )

            managers[manager_name] = manager.get_formatted_manager_name()

        return managers



from typing import Optional, Tuple

MAX_INTERFACE_NAME_LENGTH = 15


class ExternalLink(object):
    """A Kathara external collision domain. It is used to create a collision domain attached to a host interface."""

    __slots__ = ['interface', 'vlan']

    def __init__(self, interface: str, vlan: Optional[int] = None) -> None:
        self.interface: str = interface
        self.vlan: int = vlan

    def get_name_and_vlan(self) -> Tuple[str, Optional[int]]:
        """Return a tuple composed by the name of the attached interface and, if present, the vlan tag.

        The interface name is computed appending the interface name to the vlan tag (if present).
        If the length of interface name + vlan tag is more than 15 chars, the interface name is truncated in order
        to fit the whole string in 15 chars (due to Linux limitations).

        Returns:
            Tuple[str, Optional[int]]: A tuple composed by the name of the attached interface and the vlan tag.
        """
        # VLAN is defined
        if self.vlan:
            vlan_name_length = len(".%s" % self.vlan)

            # If the length of interface name + vlan tag is more than 15 chars, we truncate the interface name to
            # 15 - VLAN_NAME_LENGTH in order to fit the whole string in 15 chars
            return (self.interface, self.vlan) if len(self.interface) + vlan_name_length <= MAX_INTERFACE_NAME_LENGTH \
                else (self.interface[0:(MAX_INTERFACE_NAME_LENGTH - vlan_name_length)], self.vlan)

        return self.interface, None

    def get_full_name(self) -> str:
        """Return the external collision domain full name in the format: |name|.|vlan_tag|.

        Returns:
            (str): The external collision domain full name
        """
        (name, vlan) = self.get_name_and_vlan()
        return name if not vlan else "%s.%s" % (name, vlan)

    def __repr__(self) -> str:
        return "ExternalLink(%s, %s)" % (self.interface, self.vlan)



from __future__ import annotations

import logging
import os


class PrivilegeHandler(object):
    __slots__ = ['user_uid', 'user_gid', 'effective_user_uid', 'effective_user_gid']

    __instance: PrivilegeHandler = None

    @staticmethod
    def get_instance() -> PrivilegeHandler:
        if PrivilegeHandler.__instance is None:
            PrivilegeHandler()

        return PrivilegeHandler.__instance

    def __init__(self) -> None:
        if PrivilegeHandler.__instance is not None:
            raise Exception("This class is a singleton!")
        else:
            try:
                self.user_uid: int = os.getuid()
                self.user_gid: int = os.getgid()

                self.effective_user_uid: int = os.geteuid()
                self.effective_user_gid: int = os.getegid()
            except AttributeError:
                pass

            PrivilegeHandler.__instance = self

    def drop_privileges(self) -> None:
        logging.debug("Dropping privileges to UID=%d and GID=%d..." % (self.user_uid, self.user_gid))

        try:
            os.setuid(self.user_uid)
        except OSError:
            pass

        try:
            os.setgid(self.user_gid)
        except OSError:
            pass

    def raise_privileges(self) -> None:
        logging.debug("Raising privileges to UID=%d and GID=%d..." % (self.effective_user_uid, self.effective_user_gid))

        try:
            os.setuid(self.effective_user_uid)
        except OSError:
            pass

        try:
            os.setgid(self.effective_user_gid)
        except OSError:
            pass



from . import utils as setting_utils
from ....foundation.cli.ui.setting.OptionsHandler import OptionsHandler
from ....setting.addon.DockerSettingsAddon import DEFAULTS
from ....trdparty.consolemenu import *
from ....trdparty.consolemenu.items import *
from ....trdparty.consolemenu.validators.regex import RegexValidator
from ....utils import exec_by_platform


class DockerOptionsHandler(OptionsHandler):
    def add_items(self, current_menu: ConsoleMenu, menu_formatter: MenuFormatBuilder) -> None:
        # Hosthome Mount Option
        hosthome_string = "Automatically mount /hosthome on startup"
        hosthome_menu = SelectionMenu(strings=[],
                                      title=hosthome_string,
                                      subtitle=setting_utils.current_bool("hosthome_mount"),
                                      prologue_text="""The home directory of the current user is made available for """
                                                    """reading/writing inside the device under the special """
                                                    """directory `/hosthome`.
                                                    
                                                    Default is %s.""" %
                                                    setting_utils.format_bool(DEFAULTS['hosthome_mount']),
                                      formatter=menu_formatter
                                      )

        hosthome_menu.append_item(FunctionItem(text="Yes",
                                               function=setting_utils.update_setting_value,
                                               args=["hosthome_mount", True],
                                               should_exit=True
                                               )
                                  )
        hosthome_menu.append_item(FunctionItem(text="No",
                                               function=setting_utils.update_setting_value,
                                               args=["hosthome_mount", False],
                                               should_exit=True
                                               )
                                  )

        hosthome_item = SubmenuItem(hosthome_string, hosthome_menu, current_menu)

        # Shared Mount Option
        shared_string = "Automatically mount /shared on startup"
        shared_menu = SelectionMenu(strings=[],
                                    title=shared_string,
                                    subtitle=setting_utils.current_bool("shared_mount"),
                                    prologue_text="""The shared directory inside the lab folder is made available """
                                                  """for reading/writing inside the device under the special """
                                                  """directory `/shared`.
                                                  
                                                  Default is %s.""" %
                                                  setting_utils.format_bool(DEFAULTS['shared_mount']),
                                    formatter=menu_formatter
                                    )

        shared_menu.append_item(FunctionItem(text="Yes",
                                             function=setting_utils.update_setting_value,
                                             args=["shared_mount", True],
                                             should_exit=True
                                             )
                                )
        shared_menu.append_item(FunctionItem(text="No",
                                             function=setting_utils.update_setting_value,
                                             args=["shared_mount", False],
                                             should_exit=True
                                             )
                                )

        shared_item = SubmenuItem(shared_string, shared_menu, current_menu)

        # Shared Links Option
        shared_cd_string = "Enable Shared Collision Domains between users"
        shared_cd_menu = SelectionMenu(strings=[],
                                       title=shared_cd_string,
                                       subtitle=setting_utils.current_bool("shared_cd"),
                                       prologue_text="""This option allows to connect devices of different users to """
                                                     """the same collision domains.
                                                     
                                                     Default is %s.""" %
                                                     setting_utils.format_bool(DEFAULTS['shared_cd']),
                                       formatter=menu_formatter
                                       )

        shared_cd_menu.append_item(FunctionItem(text="Yes",
                                                function=setting_utils.update_setting_value,
                                                args=["shared_cd", True],
                                                should_exit=True
                                                )
                                   )
        shared_cd_menu.append_item(FunctionItem(text="No",
                                                function=setting_utils.update_setting_value,
                                                args=["shared_cd", False],
                                                should_exit=True
                                                )
                                   )

        shared_cd_item = SubmenuItem(shared_cd_string, shared_cd_menu, current_menu)

        # Remote Docker Daemon Option
        def remote_url_unix():
            remote_url_string = "Configure a remote Docker connection"
            remote_url_menu = SelectionMenu(strings=[],
                                            title=remote_url_string,
                                            subtitle=setting_utils.current_string("remote_url"),
                                            prologue_text="""You can specify a remote Docker Daemon URL.
                                            Default is %s.""" % DEFAULTS['remote_url'],
                                            formatter=menu_formatter
                                            )

            remote_url_menu.append_item(FunctionItem(text="Insert a remote Docker Daemon URL",
                                                     function=setting_utils.read_value,
                                                     args=['remote_url',
                                                           RegexValidator(setting_utils.URL_REGEX),
                                                           'Write a Docker Daemon URL '
                                                           '(format http[s]://<remote-url>:<remote-port>):',
                                                           'Docker Daemon URL is not a valid URL (remove '
                                                           'the trailing slash, if present)'
                                                           ],
                                                     should_exit=False
                                                     )
                                        )

            remote_url_item = SubmenuItem(remote_url_string, remote_url_menu, current_menu)

            # Docker Daemon TLS Path Option
            cert_path_string = "Configure a Docker Daemon TLS Cert Path"
            cert_path_menu = SelectionMenu(strings=[],
                                           title=cert_path_string,
                                           subtitle=setting_utils.current_string("cert_path"),
                                           prologue_text="""When using a remote Docker Daemon, a TLS Cert could be required.
                                           Default is %s.""" % DEFAULTS['cert_path'],
                                           formatter=menu_formatter
                                           )

            cert_path_menu.append_item(FunctionItem(text="Insert a Docker Daemon TLS Cert Path",
                                                    function=setting_utils.read_value,
                                                    args=['cert_path',
                                                          RegexValidator(r'^.+$'),
                                                          'Write a TSL Cert Path:',
                                                          'TLS Cert Path not valid!'
                                                          ],
                                                    should_exit=False
                                                    )
                                       )
            cert_path_menu.append_item(FunctionItem(text="Reset value to Empty String",
                                                    function=setting_utils.update_setting_value,
                                                    args=["cert_path", None],
                                                    should_exit=False
                                                    )
                                       )

            cert_path_item = SubmenuItem(cert_path_string, cert_path_menu, remote_url_menu)
            remote_url_menu.append_item(cert_path_item)

            remote_url_menu.append_item(FunctionItem(text="Reset remote Docker connection to default",
                                                     function=setting_utils.update_setting_values,
                                                     args=[[("remote_url", None), ("cert_path", None)]],
                                                     should_exit=False
                                                     )
                                        )

            return remote_url_item

        remote_url_item = exec_by_platform(remote_url_unix, lambda: None, remote_url_unix)

        current_menu.append_item(hosthome_item)
        current_menu.append_item(shared_item)
        current_menu.append_item(shared_cd_item)
        if remote_url_item:
            current_menu.append_item(remote_url_item)



import argparse
from typing import List

from ... import utils
from ...foundation.cli.command.Command import Command
from ...manager.Kathara import Kathara
from ...strings import strings, wiki_description


class LcleanCommand(Command):
    def __init__(self) -> None:
        Command.__init__(self)

        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(
            prog='kathara lclean',
            description=strings['lclean'],
            epilog=wiki_description,
            add_help=False
        )

        self.parser.add_argument(
            '-h', '--help',
            action='help',
            default=argparse.SUPPRESS,
            help='Show an help message and exit.'
        )

        self.parser.add_argument(
            '-d', '--directory',
            required=False,
            help='Specify the folder containing the lab.'
        )
        self.parser.add_argument(
            'machine_names',
            metavar='DEVICE_NAME',
            nargs='*',
            help='Clean only specified devices.'
        )

    def run(self, current_path: str, argv: List[str]) -> None:
        self.parse_args(argv)
        args = self.get_args()

        lab_path = args['directory'].replace('"', '').replace("'", '') if args['directory'] else current_path
        lab_path = utils.get_absolute_path(lab_path)

        lab_hash = utils.generate_urlsafe_hash(lab_path)

        Kathara.get_instance().undeploy_lab(lab_hash,
                                            selected_machines=set(args['machine_names'])
                                            )



import os
from abc import ABC, abstractmethod
from typing import Dict, Union, List

from deepdiff import DeepDiff

from ...manager.Kathara import Kathara
from ...model.Lab import Lab


class Test(ABC):
    __slots__ = ['lab', 'test_path', 'signature_path', 'results_path']

    def __init__(self, lab: Lab) -> None:
        self.lab: Lab = lab
        self.test_path: str = os.path.join(self.lab.path, '_test')

        self.signature_path: str = os.path.join(self.test_path, 'signature')
        self.results_path: str = os.path.join(self.test_path, 'results')

    @abstractmethod
    def create_signature(self) -> None:
        raise NotImplementedError("You must implement `create_signature` method.")

    @abstractmethod
    def test(self) -> bool:
        raise NotImplementedError("You must implement `test` method.")

    @abstractmethod
    def check_signature(self, signature: Union[Dict, str], status: Union[Dict, str]) -> Union[DeepDiff, List]:
        raise NotImplementedError("You must implement `check_signature` method.")

    @staticmethod
    def _get_machine_command_output(lab_hash, machine_name, command):
        exec_output = Kathara.get_instance().exec(lab_hash, machine_name, command)

        result = {
            'stdout': '',
            'stderr': ''
        }

        try:
            while True:
                (stdout, stderr) = next(exec_output)
                if stdout:
                    result['stdout'] += stdout.decode('utf-8')
                if stderr:
                    result['stderr'] += stderr.decode('utf-8')
        except StopIteration:
            return result['stdout'], result['stderr']



import hashlib
import logging
import re
from functools import partial
from multiprocessing import Manager
from multiprocessing.dummy import Pool
from typing import Dict, Optional, Set, Any, List

import progressbar
from kubernetes import client
from kubernetes.client.api import custom_objects_api
from kubernetes.client.rest import ApiException

from .KubernetesConfig import KubernetesConfig
from ... import utils
from ...model.Lab import Lab
from ...model.Link import Link
from ...setting.Setting import Setting

MAX_K8S_LINK_NUMBER = (1 << 24) - 10

K8S_NET_GROUP = "k8s.cni.cncf.io"
K8S_NET_VERSION = "v1"
K8S_NET_PLURAL = "network-attachment-definitions"


class KubernetesLink(object):
    """Class responsible for interacting with Kubernetes."""
    __slots__ = ['client', 'seed']

    def __init__(self) -> None:
        self.client: custom_objects_api.CustomObjectsApi = custom_objects_api.CustomObjectsApi()

        self.seed: str = KubernetesConfig.get_cluster_user()

    def deploy_links(self, lab: Lab, selected_links: Dict[str, Link] = None) -> None:
        """Deploy all the links contained in lab.links.

        Args:
            lab (Kathara.model.Lab.Lab): A Kathara network scenario.
            selected_links (Dict[str, Link]): Keys are Link names, values are Link objects.

        Returns:
            None
        """
        links = selected_links.items() if selected_links else lab.links.items()

        if len(links) > 0:
            pool_size = utils.get_pool_size()
            link_pool = Pool(pool_size)

            items = utils.chunk_list(links, pool_size)

            progress_bar = None
            if utils.CLI_ENV:
                progress_bar = progressbar.ProgressBar(
                    widgets=['Deploying collision domains... ', progressbar.Bar(),
                             ' ', progressbar.Counter(format='%(value)d/%(max_value)d')],
                    redirect_stdout=True,
                    max_value=len(links)
                )

            with Manager() as manager:
                network_ids = manager.dict()

                for chunk in items:
                    link_pool.map(func=partial(self._deploy_link, progress_bar, network_ids), iterable=chunk)

            if utils.CLI_ENV:
                progress_bar.finish()

    def _deploy_link(self, progress_bar: progressbar.ProgressBar, network_ids: Dict, link_item: (str, Link)) -> None:
        """Deploy the Link contained in the link_item.

        Args:
            progress_bar (Optional[progressbar.ProgressBar]): A progress bar object to display if used from cli.
            network_ids (Dict):
            link_item (Tuple[str, Link]): A tuple composed by the name of the link and a Link object

        Returns:
            None
        """
        (_, link) = link_item

        network_id = self._get_unique_network_id(link.name, network_ids)
        self.create(link, network_id)

        if progress_bar is not None:
            progress_bar += 1

    def create(self, link: Link, network_id: int) -> None:
        """Create a Docker Network representing the link object and assign it to link.api_object.

        Args:
            link (Kathara.model.Link.Link): A Kathara collision domain.
            network_id (int): The Network ID.

        Returns:
            None
        """
        # If a network with the same name exists, return it instead of creating a new one.
        network_objects = self.get_links_api_objects_by_filters(lab_hash=link.lab.hash, link_name=link.name)
        if network_objects:
            link.api_object = network_objects.pop()
            return

        link.api_object = self.client.create_namespaced_custom_object(group=K8S_NET_GROUP,
                                                                      version=K8S_NET_VERSION,
                                                                      namespace=link.lab.hash,
                                                                      plural=K8S_NET_PLURAL,
                                                                      body=self._build_definition(link, network_id)
                                                                      )

        # If external is defined for a link, throw a warning.
        if link.external:
            logging.warning('External is not supported on Megalos. It will be ignored.')

    def undeploy(self, lab_hash: str, networks_to_delete: Optional[Set] = None) -> None:
        """Undeploy all the links of the scenario specified by lab_hash.

        Args:
            lab_hash (str): The hash of the network scenario to undeploy.
            networks_to_delete (Set): If specified, delete only the networks that contained.

        Returns:
            None
        """
        links = self.get_links_api_objects_by_filters(lab_hash=lab_hash)
        if networks_to_delete is not None and len(networks_to_delete) > 0:
            links = [item for item in links if item["metadata"]["name"] in networks_to_delete]

        if len(links) > 0:
            pool_size = utils.get_pool_size()
            links_pool = Pool(pool_size)

            items = utils.chunk_list(links, pool_size)

            progress_bar = None
            if utils.CLI_ENV:
                progress_bar = progressbar.ProgressBar(
                    widgets=['Deleting collision domains... ', progressbar.Bar(),
                             ' ', progressbar.Counter(format='%(value)d/%(max_value)d')],
                    redirect_stdout=True,
                    max_value=len(links)
                )

            for chunk in items:
                links_pool.map(func=partial(self._undeploy_link, progress_bar), iterable=chunk)

            if utils.CLI_ENV:
                progress_bar.finish()

    def wipe(self) -> None:
        """Undeploy all the running networks of the specified user.

        Returns:
            None
        """
        links = self.get_links_api_objects_by_filters()

        pool_size = utils.get_pool_size()
        links_pool = Pool(pool_size)

        items = utils.chunk_list(links, pool_size)

        for chunk in items:
            links_pool.map(func=partial(self._undeploy_link, None), iterable=chunk)

    def _undeploy_link(self, progress_bar: progressbar.ProgressBar, link_item: Any) -> None:
        """Undeploy a Kubernetes network.

        Args:
            progress_bar (Optional[progressbar.ProgressBar]): A progress bar object to display if used from cli.
            link_item (): The Kubernetes network to undeploy.

        Returns:
            None
        """
        namespace = link_item["metadata"]["namespace"]

        try:
            self.client.delete_namespaced_custom_object(group=K8S_NET_GROUP,
                                                        version=K8S_NET_VERSION,
                                                        namespace=namespace,
                                                        plural=K8S_NET_PLURAL,
                                                        name=link_item["metadata"]["name"],
                                                        body=client.V1DeleteOptions(grace_period_seconds=0),
                                                        grace_period_seconds=0
                                                        )
        except ApiException:
            pass

        if progress_bar is not None:
            progress_bar += 1

    def get_links_api_objects_by_filters(self, lab_hash: str = None, link_name: str = None) -> List[Any]:
        """Return the List of Kubernetes networks.

        Args:
            lab_hash (str): The hash of a network scenario. If specified, return only the networks in the scenario, else
            return the networks in the 'default' namespace.
            link_name (str): The name of a network. If specified, return the specified networks of the scenario.

        Returns:
            List[Any]: A list of Kubernetes networks.
        """
        filters = ["app=kathara"]
        if link_name:
            filters.append("name=%s" % link_name)

        return self.client.list_namespaced_custom_object(group=K8S_NET_GROUP,
                                                         version=K8S_NET_VERSION,
                                                         namespace=lab_hash if lab_hash else "default",
                                                         plural=K8S_NET_PLURAL,
                                                         label_selector=",".join(filters),
                                                         timeout_seconds=9999
                                                         )["items"]

    def _build_definition(self, link: Link, network_id: int) -> Dict[str, str]:
        """Return a Dict containing the network definition for Kubernetes API corresponding to link.

        Args:
            link (Kathara.model.Link.Link): A Kathara collision domain.
            network_id (int): The network id.

        Returns:
            Dict[str, str]: a Dict containing the network definition for Kubernetes API
        """
        return {
            "apiVersion": "k8s.cni.cncf.io/v1",
            "kind": "NetworkAttachmentDefinition",
            "metadata": {
                "name": self.get_network_name(link.name),
                "labels": {
                    "name": link.name,
                    "app": "kathara"
                }
            },
            "spec": {
                "config": """{
                            "cniVersion": "0.3.0",
                            "name": "%s",
                            "type": "megalos",
                            "suffix": "%s",
                            "vxlanId": %d
                        }""" % (link.name.lower(), link.lab.hash[0:6], network_id)
            }
        }

    def _get_unique_network_id(self, name: str, network_ids: Dict[int, int]) -> int:
        """Return a unique network id.

        Args:
            name (str): The name of the network.
            network_ids (Dict[int, int]): A Dict for reserving network IDs. The Key is the network_id, value is 1 if it
            is currently used.

        Returns:
            int: A network ID.

        """
        network_id = self._get_network_id(name)
        offset = 1
        while network_id in network_ids:
            network_id = self._get_network_id(name, offset)
            offset += 1

        network_ids[network_id] = 1

        return network_id

    def _get_network_id(self, name: str, offset: int = 0) -> int:
        """Return the Kubernetes network ID from a Kathara collision domain name.

        Args:
            name (str): The name of a Kathara collision domain.
            offset (int): Offset to avoid collision on the generated Kubernetes network ID.

        Returns:
            int: The Kubernetes network ID.
        """
        network_name = self.seed + name
        return (offset + int(hashlib.sha256(network_name.encode('utf-8')).hexdigest(), 16)) % MAX_K8S_LINK_NUMBER

    @staticmethod
    def get_network_name(name: str) -> str:
        """Return the name of a Kubernetes Network.

        Args:
            name (str): The name of a Kathara Link.

        Returns:
            str: The name of the Kubernetes Network in the format "<net_prefix>-<name><suffix>".
        """
        suffix = ''
        # Underscore is replaced with -, but to keep name uniqueness append 8 chars of hash from the original name
        if '_' in name:
            suffix = '-%s' % hashlib.md5(name.encode('utf-8', errors='ignore')).hexdigest()[:8]
            name = name.replace('_', '-')

        link_name = "%s-%s%s" % (Setting.get_instance().net_prefix, name, suffix)
        return re.sub(r'[^0-9a-z\-.]+', '', link_name.lower())



import io
import logging
from copy import copy
from datetime import datetime
from typing import Set, Dict, Generator, Any, Tuple, List

import docker
import docker.models.containers
from requests.exceptions import ConnectionError as RequestsConnectionError
from terminaltables import DoubleTable

from .DockerImage import DockerImage
from .DockerLink import DockerLink
from .DockerMachine import DockerMachine
from .DockerPlugin import DockerPlugin
from ... import utils
from ...decorators import privileged
from ...exceptions import DockerDaemonConnectionError
from ...foundation.manager.IManager import IManager
from ...model.Lab import Lab
from ...model.Link import BRIDGE_LINK_NAME
from ...model.Machine import Machine
from ...setting.Setting import Setting
from ...utils import pack_files_for_tar


def pywin_import_stub():
    """Stub module of pywintypes for Unix systems (so it won't raise any `module not found` exception)."""
    import types
    pywintypes = types.ModuleType("pywintypes")
    pywintypes.error = RequestsConnectionError
    return pywintypes


def pywin_import_win():
    import pywintypes
    return pywintypes


def check_docker_status(method):
    """Decorator function to check if Docker daemon is running properly."""
    pywintypes = utils.exec_by_platform(pywin_import_stub, pywin_import_win, pywin_import_stub)

    @privileged
    def check_docker(*args, **kw):
        # Call the constructor first
        method(*args, **kw)

        # Client is initialized after constructor call
        client = args[0].client

        # Try to ping Docker, to see if it's running and raise an exception on failure
        try:
            client.ping()
        except RequestsConnectionError as e:
            raise DockerDaemonConnectionError("Can not connect to Docker Daemon. %s" % str(e))
        except pywintypes.error as e:
            raise DockerDaemonConnectionError("Can not connect to Docker Daemon. %s" % str(e))

    return check_docker


class DockerManager(IManager):
    """The class responsible to interact between Kathara and the Docker APIs."""
    __slots__ = ['client', 'docker_image', 'docker_machine', 'docker_link']

    @check_docker_status
    def __init__(self) -> None:
        remote_url = Setting.get_instance().remote_url
        if remote_url is None:
            self.client: docker.DockerClient = docker.from_env(timeout=None, max_pool_size=utils.get_pool_size())
        else:
            tls_config = docker.tls.TLSConfig(ca_cert=Setting.get_instance().cert_path)
            self.client: docker.DockerClient = docker.DockerClient(base_url=remote_url, timeout=None,
                                                                   max_pool_size=utils.get_pool_size(),
                                                                   tls=tls_config)

        docker_plugin = DockerPlugin(self.client)
        docker_plugin.check_and_download_plugin()

        self.docker_image: DockerImage = DockerImage(self.client)

        self.docker_machine: DockerMachine = DockerMachine(self.client, self.docker_image)
        self.docker_link: DockerLink = DockerLink(self.client)

    @privileged
    def deploy_lab(self, lab: Lab, selected_machines: Set[str] = None) -> None:
        """Deploy a Kathara network scenario.

        Args:
            lab (Kathara.model.Lab): A Kathara network scenario.
            selected_machines (Set[str]): If not None, deploy only the specified devices.

        Returns:
            None
        """
        selected_links = None
        if selected_machines:
            selected_links = lab.get_links_from_machines(selected_machines)

        # Deploy all lab links.
        self.docker_link.deploy_links(lab, selected_links=selected_links)

        # Deploy all lab machines.
        self.docker_machine.deploy_machines(lab, selected_machines=selected_machines)

    @privileged
    def update_lab(self, lab: Lab) -> None:
        """Update a running network scenario.

        Args:
            lab (Kathara.model.Lab): A Kathara network scenario.

        Returns:
            None
        """
        # Deploy new links (if present)
        for (_, link) in lab.links.items():
            if link.name == BRIDGE_LINK_NAME:
                continue

            self.docker_link.create(link)

        # Update lab devices.
        for (_, machine) in lab.machines.items():
            # Device is not deployed, deploy it
            if machine.api_object is None:
                self.deploy_lab(lab, selected_machines={machine.name})
            else:
                # Device already deployed, update it
                self.docker_machine.update(machine)

    @privileged
    def undeploy_lab(self, lab_hash: str, selected_machines: Set[str] = None) -> None:
        """Undeploy a Kathara network scenario.

        Args:
            lab_hash (str): The hash of the network scenario to undeploy.
            selected_machines (Set[str]): If not None, undeploy only the specified devices.

        Returns:
            None
        """
        self.docker_machine.undeploy(lab_hash, selected_machines=selected_machines)

        self.docker_link.undeploy(lab_hash)

    @privileged
    def wipe(self, all_users: bool = False) -> None:
        """Undeploy all the running network scenarios.

        If multiuser scenarios are active, undeploy only current user devices.

        Args:
        all_users (bool): If false, undeploy only the current user network scenarios. If true, undeploy the
           running network scenarios of all users.

        Returns:
            None
        """
        if Setting.get_instance().remote_url is not None and all_users:
            all_users = False
            logging.warning("Cannot wipe devices of other users with a remote Docker connection.")

        user_name = utils.get_current_user_name() if not all_users else None

        self.docker_machine.wipe(user=user_name)
        self.docker_link.wipe(user=user_name)

    @privileged
    def connect_tty(self, lab_hash: str, machine_name: str, shell: str = None, logs: bool = False) -> None:
        """Connect to a device in a running network scenario, using the specified shell.

        Args:
            lab_hash (str): The hash of the network scenario to undeploy.
            machine_name (str): The name of the device to connect.
            shell (str): The name of the shell to use for connecting.
            logs (bool): If True, print startup logs on stdout.

        Returns:
            None
        """
        user_name = utils.get_current_user_name()

        self.docker_machine.connect(lab_hash=lab_hash,
                                    machine_name=machine_name,
                                    user=user_name,
                                    shell=shell,
                                    logs=logs
                                    )

    @privileged
    def exec(self, lab_hash: str, machine_name: str, command: str) -> Generator[Tuple[bytes, bytes], None, None]:
        """Exec a command on a device in a running network scenario.

        Args:
            lab_hash (str): The hash of the network scenario to undeploy.
            machine_name (str): The name of the device to connect.
            command (str): The command to exec on the device.

        Returns:
            Generator[Tuple[bytes, bytes]]: A generator of tuples containing the stdout and stderr in bytes.
        """
        user_name = utils.get_current_user_name()

        return self.docker_machine.exec(lab_hash, machine_name, command, user=user_name, tty=False)

    @privileged
    def copy_files(self, machine: Machine, guest_to_host: Dict[str, io.IOBase]) -> None:
        """Copy files on a running device in the specified paths.

        Args:
            machine (Kathara.model.Machine): A running machine object. It must have the api_object field populated.
            guest_to_host (Dict[str, io.IOBase]): A dict containing the device path as key and
                fileobj to copy in path as value.

        Returns:
            None
        """
        tar_data = pack_files_for_tar(guest_to_host)

        self.docker_machine.copy_files(machine.api_object,
                                       path="/",
                                       tar_data=tar_data
                                       )

    @privileged
    def get_lab_info(self, lab_hash: str = None, machine_name: str = None, all_users: bool = False) -> \
            Generator[Dict[str, Any], None, None]:
        """Return information about the running devices.

        Args:
            lab_hash (str): If not None, return information of the corresponding network scenario.
            machine_name (str): If not None, return information of the specified device.
            all_users (bool): If True, return information about the device of all users.

        Returns:
              Generator[Dict[str, Any], None, None]: A generator containing dicts containing device names as keys and
              their info as values.
        """
        user_name = utils.get_current_user_name() if not all_users else None

        lab_info = self.docker_machine.get_machines_info(lab_hash, machine_name=machine_name, user=user_name)

        return lab_info

    @privileged
    def get_formatted_lab_info(self, lab_hash: str = None, machine_name: str = None, all_users: bool = False) -> str:
        """Return a formatted string with the information about the running devices.

        Args:
            lab_hash (str): If not None, return information of the corresponding network scenario.
            machine_name (str): If not None, return information of the specified device.
            all_users (bool): If True, return information about the device of all users.

        Returns:
             str: String containing devices info
        """
        table_header = ["LAB HASH", "USER", "DEVICE NAME", "STATUS", "CPU %", "MEM USAGE / LIMIT", "MEM %", "NET I/O"]
        stats_table = DoubleTable([])
        stats_table.inner_row_border = True

        lab_info = self.get_lab_info(lab_hash, machine_name, all_users)

        while True:
            machines_data = [
                table_header
            ]

            try:
                result = next(lab_info)
            except StopIteration:
                return

            if not result:
                return

            for machine_stats in result:
                machines_data.append([machine_stats['real_lab_hash'],
                                      machine_stats['user'],
                                      machine_stats['name'],
                                      machine_stats['status'],
                                      machine_stats['cpu_usage'],
                                      machine_stats['mem_usage'],
                                      machine_stats['mem_percent'],
                                      machine_stats['net_usage']
                                      ])

            stats_table.table_data = machines_data

            yield "TIMESTAMP: %s" % datetime.now() + "\n\n" + stats_table.table

    @privileged
    def get_machine_api_object(self, lab_hash: str, machine_name: str) -> docker.models.containers.Container:
        """Return the corresponding API object of a running device in a network scenario.

        Args:
            lab_hash (str): The hash of the network scenario.
            machine_name (str): The name of the device.

        Returns:
            docker.models.containers.Container: docker machine api object.
        """
        user_name = utils.get_current_user_name()

        return self.docker_machine.get_machine_api_object(lab_hash, machine_name, user=user_name)

    @privileged
    def get_machine_info(self, machine_name: str, lab_hash: str = None, all_users: bool = False) \
            -> List[Dict[str, Any]]:
        """Return information of running devices with a specified name.

        Args:
            machine_name (str): The device name.
            lab_hash (str): If not None, search the device in the specified network scenario.
            all_users (bool): If True, search the device among all the users devices.

        Returns:
            List[Dict[str, Any]]: A list of dicts containing the devices info.
        """
        user_name = utils.get_current_user_name() if not all_users else None

        machines_stats = self.docker_machine.get_machine_info(machine_name, lab_hash=lab_hash, user=user_name)

        return machines_stats

    @privileged
    def get_formatted_machine_info(self, machine_name: str, lab_hash: str = None, all_users: bool = False) -> str:
        """Return formatted information of running devices with a specified name.

        Args:
            machine_name (str): The device name.
            lab_hash (str): If not None, search the device in the specified network scenario.
            all_users (bool): If True, search the device among all the users devices.

        Returns:
            str: The formatted devices properties.
        """
        machines_stats = self.get_machine_info(machine_name, lab_hash, all_users)

        machines_info = []

        for machine_stats in machines_stats:
            machine_info = utils.format_headers("Device information") + "\n"
            machine_info += "Lab Hash: %s\n" % machine_stats['real_lab_hash']
            machine_info += "Device Name: %s\n" % machine_stats['name']
            machine_info += "Real Device Name: %s\n" % machine_stats['real_name']
            machine_info += "Status: %s\n" % machine_stats['status']
            machine_info += "Image: %s\n\n" % machine_stats['image']
            machine_info += "PIDs: %s\n" % machine_stats['pids']
            machine_info += "CPU Usage: %s\n" % machine_stats["cpu_usage"]
            machine_info += "Memory Usage: %s\n" % machine_stats["mem_usage"]
            machine_info += "Network Usage (DL/UL): %s\n" % machine_stats["net_usage"]
            machine_info += utils.format_headers()

            machines_info.append(machine_info)

        return "\n\n".join(machines_info)

    @privileged
    def check_image(self, image_name: str) -> None:
        """Check if the specified image is valid.

        Args:
            image_name (str): The name of the image

        Returns:
            None

        Raises:
            ConnectionError: The image is not locally available and there is no connection to a remote image repository.
            Exception: The image is not found.
        """
        self.docker_image.check(image_name)

    @privileged
    def get_release_version(self) -> str:
        """Return the current manager version.

        Returns:
            str: The current manager version.
        """
        return self.client.version()["Version"]

    @staticmethod
    def get_formatted_manager_name() -> str:
        """Return a formatted string containing the current manager name.

        Returns:
            str: A formatted string containing the current manager name.
        """
        return "Docker (Kathara)"



import textwrap

from .format import MenuStyle


class Dimension(object):
    """
    The Dimension class encapsulates the height and width of a component.

    Args:
        width (int): the width of the Dimension, in columns.
        height (int): the height of the Dimension, in rows.
        dimension (Dimension, optional): an existing Dimension from which to duplicate the height and width.
    """

    def __init__(self, width=0, height=0, dimension=None):
        self.width = width
        self.height = height
        if dimension is not None:
            self.width = dimension.width
            self.height = dimension.height


class MenuComponent(object):
    """
    Base class for a menu component.

    Args:
        menu_style (:obj:`MenuStyle`): the style for this component.
        max_dimension (:obj:`Dimension`): the maximum Dimension (width x height) for the menu. Defaults to width=80
            and height=40 if not specified.

    Raises:
        TypeError: if menu_style is not a :obj:`MenuStyle`.
    """

    def __init__(self, menu_style, max_dimension=None):
        if not isinstance(menu_style, MenuStyle):
            raise TypeError('menu_style must be of type MenuStyle')
        if max_dimension is None:
            max_dimension = Dimension(width=80, height=40)
        self.__max_dimension = max_dimension
        self.__style = menu_style

    @property
    def max_dimension(self):
        """
        :obj:`Dimension`: The maximum dimension for the menu.
        """
        return self.__max_dimension

    @property
    def style(self):
        """
        :obj:`consolemenu.format.MenuStyle`: The style for this component.
        """
        return self.__style

    @property
    def margins(self):
        """
        :obj:`consolemenu.format.MenuMargins`: The margins for this component.
        """
        return self.__style.margins

    @property
    def padding(self):
        """
        :obj:`consolemenu.format.MenuPadding`: The padding for this component.
        """
        return self.__style.padding

    @property
    def border_style(self):
        """
        :obj:`consolemenu.format.MenuBorderStyle`: The border style for this component.
        """
        return self.__style.border_style

    def calculate_border_width(self):
        """
        Calculate the width of the menu border. This will be the width of the maximum allowable
        dimensions (usually the screen size), minus the left and right margins and the newline character.
        For example, given a maximum width of 80 characters, with left and right margins both
        set to 1, the border width would be 77 (80 - 1 - 1 - 1 = 77).

        Returns:
            int: the menu border width in columns.
        """
        return self.max_dimension.width - self.margins.left - self.margins.right - 1  # 1=newline

    def calculate_content_width(self):
        """
        Calculate the width of inner content of the border.  This will be the width of the menu borders,
        minus the left and right padding, and minus the two vertical border characters.
        For example, given a border width of 77, with left and right margins each set to 2, the content
        width would be 71 (77 - 2 - 2 - 2 = 71).

        Returns:
            int: the inner content width in columns.
        """
        return self.calculate_border_width() - self.padding.left - self.padding.right - 2

    def generate(self):
        """
        Generate this component.

        Yields:
            str: The next string of characters for drawing this component.
        """
        raise NotImplemented()

    def inner_horizontals(self):
        """
        The string of inner horizontal border characters of the required length for this component (not including
        the menu margins or verticals).

        Returns:
            str: The inner horizontal characters.
        """
        return u"{0}".format(self.border_style.inner_horizontal * (self.calculate_border_width() - 2))

    def inner_horizontal_border(self):
        """
        The complete inner horizontal border section, including the left and right border verticals.

        Returns:
            str: The complete inner horizontal border.
        """
        return u"{lm}{lv}{hz}{rv}".format(lm=' ' * self.margins.left,
                                          lv=self.border_style.outer_vertical_inner_right,
                                          rv=self.border_style.outer_vertical_inner_left,
                                          hz=self.inner_horizontals())

    def outer_horizontals(self):
        """
        The string of outer horizontal border characters of the required length for this component (not including
        the menu margins or verticals).

        Returns:
            str: The outer horizontal characters.
        """
        return u"{0}".format(self.border_style.outer_horizontal * (self.calculate_border_width() - 2))

    def outer_horizontal_border_bottom(self):
        """
        The complete outer bottom horizontal border section, including left and right margins.

        Returns:
            str: The bottom menu border.
        """
        return u"{lm}{lv}{hz}{rv}".format(lm=' ' * self.margins.left,
                                          lv=self.border_style.bottom_left_corner,
                                          rv=self.border_style.bottom_right_corner,
                                          hz=self.outer_horizontals())

    def outer_horizontal_border_top(self):
        """
        The complete outer top horizontal border section, including left and right margins.

        Returns:
            str: The top menu border.
        """
        return u"{lm}{lv}{hz}{rv}".format(lm=' ' * self.margins.left,
                                          lv=self.border_style.top_left_corner,
                                          rv=self.border_style.top_right_corner,
                                          hz=self.outer_horizontals())

    def row(self, content='', align='left'):
        """
        A row of the menu, which comprises the left and right verticals plus the given content.

        Returns:
            str: A row of this menu component with the specified content.
        """
        return u"{lm}{vert}{cont}{vert}".format(lm=' ' * self.margins.left,
                                                vert=self.border_style.outer_vertical,
                                                cont=self._format_content(content, align))

    @staticmethod
    def _alignment_char(align):
        if str(align).strip() == 'center':
            return '^'
        elif str(align).strip() == 'right':
            return '>'
        else:
            return '<'

    def _format_content(self, content='', align='left'):
        return '{lp}{text:{al}{width}}{rp}'.format(lp=' ' * self.padding.left,
                                                   rp=' ' * self.padding.right,
                                                   text=content, al=self._alignment_char(align),
                                                   width=(self.calculate_border_width() - self.padding.left -
                                                          self.padding.right - 2))


class MenuHeader(MenuComponent):
    """
    The menu header section.
    The menu header contains the top margin, menu top, title/subtitle verticals, bottom padding verticals,
    and optionally a bottom border to separate the header from the next section.
    """

    def __init__(self, menu_style, max_dimension=None, title=None, title_align='left',
                 subtitle=None, subtitle_align='left', show_bottom_border=False):
        super(MenuHeader, self).__init__(menu_style, max_dimension)
        self.title = title
        self.title_align = title_align
        self.subtitle = subtitle
        self.subtitle_align = subtitle_align
        self.show_bottom_border = show_bottom_border

    def generate(self):
        for x in range(0, self.margins.top):
            yield ''
        yield self.outer_horizontal_border_top()
        for x in range(0, self.padding.top):
            yield self.row()
        if self.title is not None and self.title != '':
            yield self.row(content=self.title, align=self.title_align)
        if self.subtitle is not None and self.subtitle != '':
            yield self.row()
            yield self.row(content=self.subtitle, align=self.subtitle_align)
        for x in range(0, self.padding.bottom):
            yield self.row()
        if self.show_bottom_border:
            yield self.inner_horizontal_border()


class MenuTextSection(MenuComponent):
    """
    The menu text block section.
    A text block section can be used for displaying text to the user above or below the main items section.
    """

    def __init__(self, menu_style, max_dimension=None, text=None, text_align='left',
                 show_top_border=False, show_bottom_border=False):
        super(MenuTextSection, self).__init__(menu_style, max_dimension)
        self.text = text
        self.text_align = text_align
        self.show_top_border = show_top_border
        self.show_bottom_border = show_bottom_border

    def generate(self):
        if self.show_top_border:
            yield self.inner_horizontal_border()
        for x in range(0, self.padding.top):
            yield self.row()
        if self.text is not None and self.text != '':
            for line in textwrap.wrap(self.text, width=self.calculate_content_width()):
                yield self.row(content=line, align=self.text_align)
        for x in range(0, self.padding.bottom):
            yield self.row()
        if self.show_bottom_border:
            yield self.inner_horizontal_border()


class MenuItemsSection(MenuComponent):
    """
    The menu section for displaying the menu items.
    """

    def __init__(self, menu_style, max_dimension=None, items=None, items_align='left'):
        super(MenuItemsSection, self).__init__(menu_style, max_dimension)
        if items is not None:
            self.__items = items
        else:
            self.__items = list()
        self.items_align = items_align
        self.__top_border_dict = dict()
        self.__bottom_border_dict = dict()

    @property
    def items(self):
        return self.__items

    @items.setter
    def items(self, items):
        self.__items = items

    @property
    def items_with_bottom_border(self):
        """
        Return a list of the names (the item text property) of all items that should show a bottom border.
        :return: a list of item names that should show a bottom border.
        """
        return self.__bottom_border_dict.keys()

    @property
    def items_with_top_border(self):
        """
        Return a list of the names (the item text property) of all items that should show a top border.
        :return: a list of item names that should show a top border.
        """
        return self.__top_border_dict.keys()

    def show_item_bottom_border(self, item_text, flag):
        """
        Sets a flag that will show a bottom border for an item with the specified text.
        :param item_text: the text property of the item
        :param flag: boolean specifying if the border should be shown.
        """
        if flag:
            self.__bottom_border_dict[item_text] = True
        else:
            self.__bottom_border_dict.pop(item_text, None)

    def show_item_top_border(self, item_text, flag):
        """
        Sets a flag that will show a top border for an item with the specified text.
        :param item_text: the text property of the item
        :param flag: boolean specifying if the border should be shown.
        """
        if flag:
            self.__top_border_dict[item_text] = True
        else:
            self.__top_border_dict.pop(item_text, None)

    def generate(self):
        for x in range(0, self.padding.top):
            yield self.row()
        for index, item in enumerate(self.items):
            if item.text in self.items_with_top_border:
                yield self.inner_horizontal_border()
            yield self.row(content=item.show(index), align=self.items_align)
            if item.text in self.items_with_bottom_border:
                yield self.inner_horizontal_border()
        for x in range(0, self.padding.bottom):
            yield self.row()


class MenuFooter(MenuComponent):
    """
    The menu footer section.
    The menu footer contains the menu bottom, bottom padding verticals, and bottom margin.
    """

    def generate(self):
        for x in range(0, self.padding.top):
            yield self.row()
        yield self.outer_horizontal_border_bottom()
        for x in range(0, self.margins.bottom):
            yield ''


class MenuPrompt(MenuComponent):
    """
    A string representing the menu prompt for user input.
    """

    def __init__(self, menu_style, max_dimension=None, prompt_string=">>"):
        super(MenuPrompt, self).__init__(menu_style, max_dimension)
        self.__prompt = prompt_string

    @property
    def prompt(self):
        return self.__prompt

    @prompt.setter
    def prompt(self, prompt):
        self.__prompt = prompt

    def generate(self):
        for x in range(0, self.padding.top):
            yield ''
        for line in self.prompt.split():
            yield u"{lm}{line} ".format(lm=' ' * self.margins.left, line=line)



from . import utils as setting_utils
from ....foundation.cli.ui.setting.OptionsHandler import OptionsHandler
from ....setting.addon.KubernetesSettingsAddon import DEFAULTS
from ....trdparty.consolemenu import *
from ....trdparty.consolemenu.items import *
from ....trdparty.consolemenu.validators.regex import RegexValidator


class KubernetesOptionsHandler(OptionsHandler):
    def add_items(self, current_menu: ConsoleMenu, menu_formatter: MenuFormatBuilder) -> None:
        # API URL Option
        api_server_url_string = "Insert a Kubernetes API Server URL"
        api_server_url_menu = SelectionMenu(strings=[],
                                            title=api_server_url_string,
                                            subtitle=setting_utils.current_string("api_server_url"),
                                            prologue_text="""You can specify a remote Kubernetes API Server URL to """
                                                          """connect to when Megalos is not used on a Kubernetes """
                                                          """master.
                                                          Default is %s.""" % DEFAULTS['api_server_url'],
                                            formatter=menu_formatter
                                            )

        api_server_url_menu.append_item(FunctionItem(text=api_server_url_string,
                                                     function=setting_utils.read_value,
                                                     args=['api_server_url',
                                                           RegexValidator(setting_utils.URL_REGEX),
                                                           'Write a Kubernetes API Server URL:',
                                                           'Kubernetes API Server URL is not a valid URL (remove '
                                                           'the trailing slash, if present)'
                                                           ],
                                                     should_exit=True
                                                     )
                                        )
        api_server_url_menu.append_item(FunctionItem(text="Reset value to Empty String",
                                                     function=setting_utils.update_setting_value,
                                                     args=["api_server_url", None],
                                                     should_exit=True
                                                     )
                                        )

        api_url_item = SubmenuItem(api_server_url_string, api_server_url_menu, current_menu)

        # API Token Option
        api_token_string = "Insert a Kubernetes API Token"
        api_token_menu = SelectionMenu(strings=[],
                                       title=api_token_string,
                                       prologue_text="""When using a remote Kubernetes API Server, you must also """
                                                     """specify the authentication token to use.
                                                     Default is %s.""" % DEFAULTS['api_token'],
                                       formatter=menu_formatter
                                       )

        api_token_menu.append_item(FunctionItem(text=api_token_string,
                                                function=setting_utils.read_value,
                                                args=['api_token',
                                                      RegexValidator(r'^.+$'),
                                                      'Write a Kubernetes API Token:',
                                                      'Kubernetes API Token not valid!'
                                                      ],
                                                should_exit=True
                                                )
                                   )
        api_token_menu.append_item(FunctionItem(text="Reset value to Empty String",
                                                function=setting_utils.update_setting_value,
                                                args=["api_token", None],
                                                should_exit=True
                                                )
                                   )

        api_token_item = SubmenuItem(api_token_string, api_token_menu, current_menu)

        # Shared Mount Option
        host_shared_string = "Automatically mount /shared on startup"
        host_shared_menu = SelectionMenu(strings=[],
                                         title=host_shared_string,
                                         subtitle=setting_utils.current_bool("host_shared"),
                                         prologue_text="""Each Kubernetes worker node creates a /home/shared """
                                                       """directory and it is made available for reading/writing """
                                                       """inside the device under the special directory `/shared`.
                                                       Default is %s.""" %
                                                       setting_utils.format_bool(DEFAULTS['host_shared']),
                                         formatter=menu_formatter
                                         )

        host_shared_menu.append_item(FunctionItem(text="Yes",
                                                  function=setting_utils.update_setting_value,
                                                  args=["host_shared", True],
                                                  should_exit=True
                                                  )
                                     )
        host_shared_menu.append_item(FunctionItem(text="No",
                                                  function=setting_utils.update_setting_value,
                                                  args=["host_shared", False],
                                                  should_exit=True
                                                  )
                                     )

        host_shared_item = SubmenuItem(host_shared_string, host_shared_menu, current_menu)

        # Image Pull Policy Option
        image_pull_policy_string = "Image Pull Policy"
        image_pull_policy_menu = SelectionMenu(strings=[],
                                               title=image_pull_policy_string,
                                               subtitle=setting_utils.current_string("image_pull_policy"),
                                               prologue_text="""Specify the image pull policy for Docker images """
                                                             """used by devices. 
                                                             Default is %s.""" % DEFAULTS['image_pull_policy'],
                                               formatter=menu_formatter
                                               )

        image_pull_policy_menu.append_item(FunctionItem(text="Always",
                                                        function=setting_utils.update_setting_value,
                                                        args=["image_pull_policy", "Always"],
                                                        should_exit=True
                                                        )
                                           )
        image_pull_policy_menu.append_item(FunctionItem(text="If Not Present",
                                                        function=setting_utils.update_setting_value,
                                                        args=["image_pull_policy", "IfNotPresent"],
                                                        should_exit=True
                                                        )
                                           )
        image_pull_policy_menu.append_item(FunctionItem(text="Never",
                                                        function=setting_utils.update_setting_value,
                                                        args=["image_pull_policy", "Never"],
                                                        should_exit=True
                                                        )
                                           )

        image_pull_policy_item = SubmenuItem(image_pull_policy_string, image_pull_policy_menu, current_menu)

        current_menu.append_item(api_url_item)
        current_menu.append_item(api_token_item)
        current_menu.append_item(host_shared_item)
        current_menu.append_item(image_pull_policy_item)



import os
from typing import Any, Optional, Callable

import msvcrt
import pyuv
import win32pipe
from docker import DockerClient
from docker.errors import APIError

from ....foundation.manager.terminal.Terminal import Terminal
from ....foundation.manager.terminal.terminal_utils import get_terminal_size_windows


class DockerNPipeTerminal(Terminal):
    __slots__ = ['client', 'exec_id', '_check_opened_timer']

    def __init__(self, handler: Any, client: DockerClient, exec_id: str):
        super().__init__(handler)

        self.client: DockerClient = client
        self.exec_id: str = exec_id

        self._check_opened_timer: Optional[pyuv.Timer] = None

    def _start_external(self) -> None:
        terminal_fd = self._convert_pipe_to_fd()

        self._external_terminal = pyuv.Pipe(self._loop)
        self._external_terminal.open(terminal_fd)
        self._external_terminal.set_blocking(False)
        self._external_terminal.start_read(self._read_external_terminal())

        self._check_opened_timer = pyuv.Timer(self._loop)
        self._check_opened_timer.start(self._is_terminal_opened(), 0, 5)

    def _convert_pipe_to_fd(self) -> Any:
        handle_id = self.handler._handle.handle
        win32pipe.SetNamedPipeHandleState(handle_id, 0x00000001, None, None)
        handle_fd = msvcrt.open_osfhandle(handle_id, os.O_APPEND)

        return handle_fd

    def _on_close(self) -> None:
        self._check_opened_timer.close()

    def _write_on_external_terminal(self) -> Callable:
        def write_on_external_terminal(handle, data, error):
            self._external_terminal.stop_read()
            self._external_terminal.write(data)
            self._external_terminal.start_read(self._read_external_terminal())

        return write_on_external_terminal

    def _resize_terminal(self) -> None:
        w, h = get_terminal_size_windows()
        self.client.api.exec_resize(self.exec_id, height=h, width=w)

    def _is_terminal_opened(self) -> Callable:
        def is_terminal_opened(timer_handle):
            try:
                self.client.api.exec_inspect(self.exec_id)
            except APIError:
                self.close()

        return is_terminal_opened



import curses


class Curses(object):
    __slots__ = ['_window']

    __instance = None

    @staticmethod
    def get_instance():
        if Curses.__instance is None:
            Curses()

        return Curses.__instance

    def __init__(self):
        if Curses.__instance is not None:
            raise Exception("This class is a singleton!")
        else:
            self._window = None

            Curses.__instance = self

    def init_window(self, noecho=True, nocbreak=True, timeout=1000, scrollok=True, keypad=True):
        self._window = curses.initscr()
        if noecho:
            curses.noecho()
        if nocbreak:
            curses.nocbreak()
        self._window.timeout(timeout)
        self._window.scrollok(scrollok)
        self._window.keypad(keypad)

    def print_string(self, string, erase=True):
        if not self._window:
            raise Exception("Window not initialized, call init_window first.")

        if erase:
            self._window.erase()

        self._window.addstr(string)
        self._window.refresh()

        key = self._window.getch()
        if key == 3:                                # CTRL+C
            raise KeyboardInterrupt

    def close(self):
        curses.nocbreak()
        self._window.keypad(False)
        curses.echo()
        curses.endwin()

        self._window = None



from distutils.core import setup

from setuptools import find_packages

setup(
    name='kathara',
    package_dir={'': 'src'},
    packages=find_packages('src'),
    py_modules=['kathara'],
    version='3.3.2',
    license='gpl-3.0',
    description='A lightweight container based emulation system.',
    author='Kathara Framework',
    author_email='contact@kathara.org',
    url='https://www.kathara.org',
    download_url='https://github.com/KatharaFramework/Kathara/archive/refs/tags/3.3.0.tar.gz',
    keywords=['NETWORK-EMULATION', 'CONTAINERS', 'NFV'],
    install_requires=[
        "binaryornot>=0.4.4",
        "docker>=4.4.4",
        "kubernetes>=20.13.0",
        "requests>=2.22.0",
        "coloredlogs>=10.0",
        "terminaltables>=3.1.0",
        "slug>=2.0",
        "deepdiff>=4.0.9",
        "pyroute2>=0.5.19",
        "progressbar2>=1.14.0",
        "libtmux>=0.8.2; platform_system == 'darwin' or platform_system == 'linux'",
        "pyuv>=1.4.0",
        "appscript>=1.1.0; platform_system == 'darwin'",
        "pypiwin32>=223; platform_system == 'win32'",
        "windows-curses>=2.1.0; platform_system == 'win32'"
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 3.9',
    ],
)



def get_terminal_size_windows() -> (int, int):
    try:
        from ctypes import windll, create_string_buffer

        # stdin handle is -10
        # stdout handle is -11
        # stderr handle is -12

        h = windll.kernel32.GetStdHandle(-12)
        csbi = create_string_buffer(22)
        res = windll.kernel32.GetConsoleScreenBufferInfo(h, csbi)

        if res:
            import struct
            (_, _, _, _, _, left, top, right, bottom, _, _) = struct.unpack("hhhhHhhhhhh", csbi.raw)
            w = right - left + 1
            h = bottom - top + 1

            return w, h
    except Exception:
        return get_terminal_size_tput()


def get_terminal_size_tput() -> (int, int):
    try:
        import subprocess
        proc = subprocess.Popen(['tput', 'cols'], stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        output = proc.communicate(input=None)
        w = int(output[0])
        proc = subprocess.Popen(['tput', 'lines'], stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        output = proc.communicate(input=None)
        h = int(output[0])

        return w, h
    except Exception:
        return 80, 25



from typing import List

from ..ui.setting.SettingsMenuFactory import SettingsMenuFactory
from ...foundation.cli.command.Command import Command


class SettingsCommand(Command):
    def run(self, current_path: str, argv: List[str]) -> None:
        menu_factory = SettingsMenuFactory()

        menu = menu_factory.create_menu()

        menu.show()



from typing import Tuple

CURRENT_VERSION = "3.3.2"


def parse(version: str) -> Tuple:
    return tuple([int(x) for x in version.split('.')])


def less_than(version: str, other_version: str) -> bool:
    version = parse(version)
    other_version = parse(other_version)

    return version < other_version



from .CommonOptionsHandler import CommonOptionsHandler
from ....foundation.cli.ui.setting.OptionsHandlerFactory import OptionsHandlerFactory
from ....setting.Setting import Setting
from ....trdparty.consolemenu import *
from ....trdparty.consolemenu.format import MenuBorderStyleType


class SettingsMenuFactory(object):
    __slots__ = ['menu_formatter']

    def __init__(self) -> None:
        self.menu_formatter: MenuFormatBuilder = MenuFormatBuilder().set_title_align('center') \
            .set_subtitle_align('center') \
            .set_prologue_text_align('center') \
            .set_border_style_type(MenuBorderStyleType.DOUBLE_LINE_BORDER) \
            .show_prologue_top_border(True) \
            .show_prologue_bottom_border(True)

    def create_menu(self) -> ConsoleMenu:
        menu = ConsoleMenu(title="Kathara Settings",
                           prologue_text="Choose the option to change.",
                           formatter=self.menu_formatter
                           )

        common_settings = CommonOptionsHandler()
        common_settings.set_menu_factory(self)
        common_settings.add(menu, self.menu_formatter)

        manager_type = Setting.get_instance().manager_type
        manager_settings = OptionsHandlerFactory().create_instance(class_args=(manager_type.capitalize(),))
        manager_settings.set_menu_factory(self)
        manager_settings.add(menu, self.menu_formatter)

        return menu
